import pathlib as _pl
import unittest as _ut
import pytest as _pt

import logging as _log
import numpy as _np
import pandas as _pd

import optihood.IO.readers as ior
import optihood.IO.writers as wr
import optihood.energy_network as en
import optihood.entities as ent
import tests.xls_helpers as xlsh

_CSV_DIR_PATH = xlsh.ROOT_DATA_DIR / "CSVs" / "basic_example_CSVs"
_input_data_dir = xlsh.ROOT_DATA_DIR / "excels" / "basic_example"
_input_data_path = _input_data_dir / "scenario.xls"


def join_if_multiple(x):
    """Needed to overcome changes when using subsequent IO for tests."""
    if isinstance(x, str):
        x = eval(x)

    if len(x) > 1:
        return ', '.join(x)

    return x[0]


def write_expected_file(nodal_data_with_unique_labels, expected_files_path: _pl.Path) -> None:
    """Helper function to update the expected files."""
    writer = wr.ScenarioFileWriterCSV(_pl.Path("irrelevant"), "individual")
    writer.data = nodal_data_with_unique_labels
    writer._write_scenario_to_file(expected_files_path)


class TestCsvScenarioReader(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None

        network = en.EnergyNetworkClass(None)
        data = _pd.ExcelFile(str(_input_data_path))
        self.expected_nodal_data = network.get_nodal_data_from_Excel(data)
        data.close()

    def test_read(self):
        """Test to ensure new reader produces 'identical' inputs to optihood."""
        csvReader = ior.CsvScenarioReader(_CSV_DIR_PATH)
        nodal_data = csvReader.read_scenario()

        errors = []

        xlsh.check_assertion(self, errors, nodal_data.keys(), self.expected_nodal_data.keys())

        for key, df_current in nodal_data.items():
            # check_dtype=False to simplify comparison between Excel and CSV files.
            xlsh.check_dataframe_assertion(errors, df_current, self.expected_nodal_data[key], check_dtype=False)

        if errors:
            raise ExceptionGroup(f"found {len(errors)} errors", errors)

    def test_get_unique_labels(self):
        """unit test"""
        df = _pd.DataFrame({ent.CommonLabels.label: ["a", "a", "b", "c"],
                            ent.CommonLabels.building: ["1", "2", "1", "1"],
                            })
        results = ior.get_unique_labels(df)
        self.assertListEqual(results, ['a__B001', 'a__B002', 'b__B001', 'c__B001'])

    def test_get_unique_buses(self):
        """unit test"""
        df = _pd.DataFrame({ent.CommonLabels.to: ["a", "a,b", "b,d,e", "c"],
                            ent.CommonLabels.building: ["1", "2", "1", "1"],
                            })
        results = ior.get_unique_buses(df, ent.CommonLabels.to)
        self.assertListEqual(results, [['a__B001'],
                                       ['a__B002', 'b__B002'],
                                       ['b__B001', 'd__B001', 'e__B001'],
                                       ['c__B001']])

    def test_add_unique_label_columns(self):
        """end2end test
        Unfortunately, the reader adds quotes to the entries.
        This leads to very strange behavior in assert_frame_equal:
        At positional index 0, first diff: ['electricityProdBus__B001'] != ['electricityProdBus__B001']
        Thus, this test undoes the work of putting strings in lists, to put them back to a single string again.
        """
        csvReader = ior.CsvScenarioReader(_CSV_DIR_PATH)
        nodal_data = csvReader.read_scenario()

        nodal_data_with_unique_labels = ior.add_unique_label_columns(nodal_data)

        expected_files_path = _pl.Path(__file__).parent / "expected_files" / "without_building_csvs"
        # =================================
        # TODO: Use this to update the expected files.  # pylint: disable=fixme
        # write_expected_file(nodal_data_with_unique_labels, expected_files_path)
        # =================================

        csvReader = ior.CsvScenarioReader(expected_files_path)
        expected_data = csvReader.read_scenario()

        errors = []

        xlsh.check_assertion(self, errors, nodal_data_with_unique_labels.keys(), expected_data.keys())

        for key, df_current in nodal_data_with_unique_labels.items():
            # check_dtype=False to simplify comparison between Excel and CSV files.
            df_expected = expected_data[key]
            from_unique = ent.CommonLabels.from_unique
            if from_unique in df_current.columns:
                df_expected[from_unique] = df_expected[from_unique].apply(join_if_multiple)
                df_current[from_unique] = df_current[from_unique].apply(join_if_multiple)

            to_unique = ent.CommonLabels.to_unique
            if to_unique in df_current.columns:
                df_expected[to_unique] = df_expected[to_unique].apply(join_if_multiple)
                df_current[to_unique] = df_current[to_unique].apply(join_if_multiple)

            connect_unique = ent.CommonLabels.connect_unique
            if connect_unique in df_current.columns:
                df_expected[connect_unique] = df_expected[connect_unique].apply(join_if_multiple)
                df_current[connect_unique] = df_current[connect_unique].apply(join_if_multiple)

            xlsh.check_dataframe_assertion(errors, df_current, df_expected, check_dtype=False)

        if errors:
            raise ExceptionGroup(f"found {len(errors)} errors", errors)

    def test_get_unique_buildings_with_circuits(self):
        """unit test"""
        df = _pd.DataFrame({ent.BuildingModelParameters.Building_Number: [1, 1, 2, 3],
                            ent.BuildingModelParameters.Circuit: [1, 2, 1, 1],
                            })
        results = ior.get_unique_buildings(df)
        self.assertListEqual(results, ['Building_model__B001_C001', 'Building_model__B001_C002',
                                       'Building_model__B002_C001', 'Building_model__B003_C001'])

    def test_get_unique_buildings_without_circuits(self):
        """unit test"""
        df = _pd.DataFrame({ent.BuildingModelParameters.Building_Number: [1, 1, 2, 3],
                            })
        results = ior.get_unique_buildings(df)
        self.assertListEqual(results, ['Building_model__B001', 'Building_model__B001', 'Building_model__B002',
                                       'Building_model__B003'])

    def test_add_unique_label_columns_including_buildings(self):
        """End2end test"""
        csvReader = ior.CsvScenarioReader(_CSV_DIR_PATH)
        nodal_data = csvReader.read_scenario()
        building_df = _pd.DataFrame({ent.BuildingModelParameters.Building_Number: [1, 1, 2, 3],
                                     ent.BuildingModelParameters.Circuit: [1, 2, 1, 1],
                                     })
        nodal_data[ent.NodeKeysOptional.building_model_parameters] = building_df

        nodal_data_with_unique_labels = ior.add_unique_label_columns(nodal_data)

        expected_files_path = _pl.Path(__file__).parent / "expected_files" / "without_building_csvs"

        # =================================
        # TODO: Use this to update the expected files.  # pylint: disable=fixme
        # write_expected_file(nodal_data_with_unique_labels, expected_files_path)
        # =================================

        csvReader = ior.CsvScenarioReader(expected_files_path)
        expected_data = csvReader.read_scenario()

        building_df_expected = building_df.copy()
        building_df_expected[ent.BuildingModelParameters.building_unique] = ['Building_model__B001_C001',
                                                                             'Building_model__B001_C002',
                                                                             'Building_model__B002_C001',
                                                                             'Building_model__B003_C001',
                                                                             ]
        expected_data[ent.NodeKeysOptional.building_model_parameters] = building_df_expected

        errors = []

        xlsh.check_assertion(self, errors, nodal_data_with_unique_labels.keys(), expected_data.keys())

        for key, df_current in nodal_data_with_unique_labels.items():
            # check_dtype=False to simplify comparison between Excel and CSV files.
            df_expected = expected_data[key]
            from_unique = ent.CommonLabels.from_unique
            if from_unique in df_current.columns:
                df_expected[from_unique] = df_expected[from_unique].apply(join_if_multiple)
                df_current[from_unique] = df_current[from_unique].apply(join_if_multiple)

            to_unique = ent.CommonLabels.to_unique
            if to_unique in df_current.columns:
                df_expected[to_unique] = df_expected[to_unique].apply(join_if_multiple)
                df_current[to_unique] = df_current[to_unique].apply(join_if_multiple)

            connect_unique = ent.CommonLabels.connect_unique
            if connect_unique in df_current.columns:
                df_expected[connect_unique] = df_expected[connect_unique].apply(join_if_multiple)
                df_current[connect_unique] = df_current[connect_unique].apply(join_if_multiple)

            xlsh.check_dataframe_assertion(errors, df_current, df_expected, check_dtype=False)

        if errors:
            raise ExceptionGroup(f"found {len(errors)} errors", errors)


class TestProfileAndOtherDataReader:
    def setUp(self):
        self.maxDiff = None

    def test_get_values_from_dataframe(self):
        """unit test"""
        df = _pd.DataFrame({"label": ["thing", "stuff"],
                            "desired_value": [1, 2],
                            })
        results = ior.ProfileAndOtherDataReader().get_values_from_dataframe(df, "stuff", "label",
                                                                            "desired_value", message="Test issue")
        assert results == 2

    def test_get_values_from_dataframe_nan_raises(self, caplog):
        """unit test"""
        df = _pd.DataFrame({"label": ["thing", "stuff"],
                            "desired_value": [1, _np.nan],
                            })
        with caplog.at_level(_log.ERROR):
            with _pt.raises(ValueError):
                ior.ProfileAndOtherDataReader().get_values_from_dataframe(df, "stuff", "label",
                                                                          "desired_value", message="Test issue")
            assert "Value empty in:" in caplog.text

    def test_get_values_from_dataframe_nan_allowed(self):
        """unit test"""
        df = _pd.DataFrame({"label": ["thing", "stuff"],
                            "desired_value": [1, _np.nan],
                            })
        results = ior.ProfileAndOtherDataReader().get_values_from_dataframe(df, "stuff", "label",
                                                                            "desired_value", message="Test issue",
                                                                            nan_allowed=True)
        assert _np.isnan(results)

    def test_get_values_from_dataframe_path(self):
        """unit test"""
        df = _pd.DataFrame({"label": ["thing", "stuff"],
                            "desired_value": [1, "path_to_profile.csv"],
                            })
        results = ior.ProfileAndOtherDataReader().get_values_from_dataframe(df, "stuff", "label",
                                                                            "desired_value", message="Test issue")
        assert results == "path_to_profile.csv"

    def test_get_values_from_dataframe_change_types(self, caplog):
        """unit test"""
        df = _pd.DataFrame({"label": ["thing", "stuff"],
                            "desired_value": [1, True],
                            })

        with caplog.at_level(_log.ERROR):
            with _pt.raises(ValueError):
                ior.ProfileAndOtherDataReader().get_values_from_dataframe(df, "stuff", "label",
                                                                          "desired_value", message="Test issue",
                                                                          desired_instances=(float))
            assert "Corrupt value in:" in caplog.text

    @_pt.mark.manual
    def test_add_weather_profiles(self):
        """Integration test without errors"""
        ior.ProfileAndOtherDataReader().add_weather_profiles()
        assert False

    def test_add_weather_profiles_raises(self, caplog):
        """unit test"""
        nodal_data = {ent.NodeKeys.profiles: _pd.DataFrame({ent.ProfileLabels.name: [ent.ProfileTypes.weather],
                                                            ent.ProfileLabels.path: ["no_such_path.no_such_suffix"],
                                                            })}

        with caplog.at_level(_log.ERROR):
            with _pt.raises(FileNotFoundError):
                ior.ProfileAndOtherDataReader().add_weather_profiles(nodal_data, {}, None)
            assert "Error in weather data file path" in caplog.text

    @_pt.mark.manual
    def test_add_electricity_cost(self):
        """Integration test without errors"""
        ior.ProfileAndOtherDataReader().add_electricity_cost()
        assert False

    def test_add_electricity_cost_raises(self, caplog):
        """unit test"""
        nodal_data = {ent.NodeKeys.commodity_sources: _pd.DataFrame(
            {ent.CommonLabels.label: [ent.CommoditySourceTypes.electricityResource],
             ent.CommoditySourcesLabels.variable_costs: ["no_such_path.no_such_suffix"],
             })}
        with caplog.at_level(_log.ERROR):
            with _pt.raises(FileNotFoundError):
                ior.ProfileAndOtherDataReader().add_electricity_cost(nodal_data, {}, None)
            assert "Error in electricity cost file path" in caplog.text

    @_pt.mark.manual
    def test_add_electricity_impact(self):
        """Integration test without errors"""
        ior.ProfileAndOtherDataReader().add_electricity_impact()
        assert False

    def test_add_electricity_impact_raises(self, caplog):
        """unit test"""
        nodal_data = {ent.NodeKeys.commodity_sources: _pd.DataFrame(
            {ent.CommonLabels.label: [ent.CommoditySourceTypes.electricityResource],
             ent.CommoditySourcesLabels.CO2_impact: ["no_such_path.no_such_suffix"],
             })}
        with caplog.at_level(_log.ERROR):
            with _pt.raises(FileNotFoundError):
                ior.ProfileAndOtherDataReader().add_electricity_impact(nodal_data, {}, None)
            assert "Error in electricity impact file path" in caplog.text

    @_pt.mark.manual
    def test_cluster_desired_column(self):
        """unit test"""
        ior.ProfileAndOtherDataReader().cluster_desired_column()
        assert False

    @_pt.mark.manual
    def test_cluster_and_multiply_desired_column(self):
        """unit test"""
        ior.ProfileAndOtherDataReader().cluster_and_multiply_desired_column()
        assert False

    @_pt.mark.manual
    def test_add_demand_profiles(self):
        """Integration test without errors"""
        ior.ProfileAndOtherDataReader().add_demand_profiles()
        assert False

    def test_add_demand_profiles_raises(self, caplog):
        """unit test"""
        nodal_data = {ent.NodeKeys.profiles: _pd.DataFrame({ent.ProfileLabels.name: [ent.ProfileTypes.demand],
                                                            ent.ProfileLabels.path: ["no_such_path.no_such_suffix"],
                                                            })}
        with caplog.at_level(_log.ERROR):
            with _pt.raises(FileNotFoundError):
                ior.ProfileAndOtherDataReader().add_demand_profiles(nodal_data, {}, 1, None)
            assert "Error in the demand profiles path" in caplog.text

    @_pt.mark.manual
    def test_maybe_add_natural_gas(self):
        """Integration test without errors"""
        ior.ProfileAndOtherDataReader().maybe_add_natural_gas()
        assert False

    def test_add_natural_gas_impact_raises(self, caplog):
        """unit test"""
        nodal_data = {ent.NodeKeys.commodity_sources: _pd.DataFrame(
            {ent.CommonLabels.label: [ent.CommoditySourceTypes.naturalGasResource],
             ent.CommoditySourcesLabels.CO2_impact: ["no_such_path.no_such_suffix"],
             })}

        with caplog.at_level(_log.ERROR):
            with _pt.raises(FileNotFoundError):
                ior.ProfileAndOtherDataReader().add_natural_gas_impact(nodal_data, {},  None)
            assert "Error in natural gas impact file path" in caplog.text

    def test_maybe_add_natural_gas_cost_raises(self, caplog):
        """unit test"""
        nodal_data = {ent.NodeKeys.commodity_sources: _pd.DataFrame(
            {ent.CommonLabels.label: [ent.CommoditySourceTypes.naturalGasResource],
             ent.CommoditySourcesLabels.variable_costs: ["no_such_path.no_such_suffix"],
             })}

        with caplog.at_level(_log.ERROR):
            with _pt.raises(FileNotFoundError):
                ior.ProfileAndOtherDataReader().add_natural_gas_costs(nodal_data, {},  None)
            assert "Error in natural gas cost file path" in caplog.text

    @_pt.mark.manual
    def test_maybe_add_building_model_with_internal_gains(self):
        """Integration test without errors"""
        ior.ProfileAndOtherDataReader().maybe_add_building_model_with_internal_gains()
        assert False

    def test_add_internal_gains_raises(self, caplog):
        """unit test"""
        nodal_data = {ent.NodeKeys.profiles: _pd.DataFrame(
            {ent.ProfileLabels.name: [ent.ProfileTypes.internal_gains],
             ent.ProfileLabels.path: ["no_such_path.no_such_suffix"],
             })}

        with caplog.at_level(_log.ERROR):
            with _pt.raises(FileNotFoundError):
                ior.ProfileAndOtherDataReader().add_internal_gains(nodal_data, {},  None)
            assert "Error in internal gains file path for the building model." in caplog.text

    def test_add_building_models_raises(self, caplog):
        """unit test"""
        nodal_data = {ent.NodeKeys.profiles: _pd.DataFrame(
            {ent.ProfileLabels.name: [ent.ProfileTypes.building_model_params],
             ent.ProfileLabels.path: ["no_such_path.no_such_suffix"],
             })}

        with caplog.at_level(_log.ERROR):
            with _pt.raises(FileNotFoundError):
                ior.ProfileAndOtherDataReader().add_building_models(nodal_data, {},  None)
            assert "Error in building model parameters file path." in caplog.text

    @_pt.mark.manual
    def test_clip_to_time_index(self):
        """unit test"""
        ior.ProfileAndOtherDataReader().clip_to_time_index()
        assert False



import pandas as _pd
import collections.abc as _abc
import dataclasses as _dc
import typing as _tp

import optihood.Visualizer.scenario_to_visualizer as _stv


@_dc.dataclass()
class EnergyNetworkGraphData:
    """ This may grow further. """
    nodes: _abc.Sequence[dict[str, dict[str, _tp.Union[str, float, int]]]]
    edges: _abc.Sequence[dict[str, dict[str, _tp.Union[str, float, int]]]]


def get_converters(initial_nodal_data: dict[str, _pd.DataFrame], nr_of_buildings: int) -> _abc.Sequence[_stv.ScenarioToVisualizerAbstract]:
    converters = []
    for sheet_name, sheet in initial_nodal_data.items():
        converter = _stv.scenario_data_factory(sheet_name)

        # ========================================================================
        """ This needs to be removed after all current sheets are implemented. """
        if not converter:
            print(f"{sheet_name} does not have a converter yet.")
            continue
        # ========================================================================
        converters += converter.set_from_dataFrame(sheet, nr_of_buildings)

    return converters


def get_graph_data(converters: _abc.Sequence[_stv.ScenarioToVisualizerAbstract]) -> EnergyNetworkGraphData:
    nodes = [converter.get_nodal_infos() for converter in converters]
    nodes = [node for node in nodes if node is not None]

    edges = []
    for converter in converters:
        edges += converter.get_edge_infos()

    return EnergyNetworkGraphData(nodes, edges)



import configparser as _cp
import dataclasses as _dc
import datetime as _dt
import logging as _log
import os as _os
import pathlib as _pl
import typing as _tp

import numpy as _np
import pandas as _pd

import optihood.entities as ent


# TODO: extract other stuff from set_using_nodal_data.
# TODO: make excelReader
# TODO: make ConfigConverter and have CsvReader and CsvWriter be standalone.
#       ConfigConverter. read, .convert, .write_to_csv, .write_to_excel.
# TODO: merge ProfileAndOtherDataReader into other readers though:
#       - Inheritance.
#       - Structure
#         CsvScenarioReader = ScenarioReaderAbstract(reader=CsvReader,
#                             ProfileAndOtherDataReader=ProfileAndOtherDataReader)


@_dc.dataclass
class CsvReader:
    """ use_function is currently implemented to deal with deprecation of a pandas feature
    (_pd.to_numeric, errors="ignore").
    This can be set to 'current' or 'future'.
    """

    dir_path: _pl.Path

    def read(self, file_name: str) -> _pd.DataFrame:
        # read_csv does not have as strong a parser as ExcelFile.parse.
        # One issue the following addresses, is the "nr as string" outputs.

        df = _pd.read_csv(self.dir_path / file_name)
        [self.make_nrs_numeric(df, column) for column in df.columns]

        return df

    @staticmethod
    def safe_to_numeric(value):
        # Suggested by pandas developers
        # https://github.com/pandas-dev/pandas/issues/59221#issuecomment-2755021659
        try:
            return _pd.to_numeric(value)
        except (ValueError, TypeError):
            return value

    @staticmethod
    def make_nrs_numeric(df: _pd.DataFrame, column_name: str) -> None:
        df[column_name] = df[column_name].apply(CsvReader.safe_to_numeric)


@_dc.dataclass
class CsvScenarioReader(CsvReader):
    """Very simplified implementation.
    Each CSV file inherently has its own data model.
    This needs to be incorporated into a validation of these inputs.
    """

    relative_file_paths: dict[str, str] = _dc.field(init=False)

    def __post_init__(self):
        paths = ent.CsvInputFilePathsRelative
        self.relative_file_paths = {
            ent.NodeKeys.buses: paths.buses,
            ent.NodeKeys.grid_connection: paths.grid_connection,
            ent.NodeKeys.commodity_sources: paths.commodity_sources,
            ent.NodeKeys.solar: paths.solar,
            ent.NodeKeys.transformers: paths.transformers,
            ent.NodeKeys.demand: paths.demand,
            ent.NodeKeys.storages: paths.storages,
            ent.NodeKeys.stratified_storage: paths.stratified_storage,
            ent.NodeKeys.profiles: paths.profiles,
            ent.NodeKeys.links: paths.links,
        }

    def read_scenario(self) -> dict[str, _pd.DataFrame]:
        # Fine-grained feedback could be provided using validators for the individual files.
        # This would include information related to the models,
        # e.g. T range should be between 7 and 12 deg C for chillers.

        data = {}
        errors = []
        for key, rel_path in self.relative_file_paths.items():
            try:
                data[key] = self.read(rel_path)
                # df_current = _pd.read_csv(path)
            except FileNotFoundError as e:
                if not key == ent.NodeKeys.links:
                    errors.append(e)

            # validation_error = self.validate(key, df_current)
            # if not validation_error:
            #     data[key] = df_current
            # else:
            #     errors.append(validation_error)

        if errors:
            # Should this be added to logging as well?
            raise ExceptionGroup("Issues with CSV files.", errors)

        return data


def parse_config(configFilePath: str):
    # Making this a class method would allow us to provide the parser.
    config = _cp.ConfigParser()
    config.read(configFilePath)
    configData = {}
    for section in config.sections():
        configData[section] = config.items(section)
    configData = {k.lower(): v for k, v in configData.items()}

    return configData


def add_unique_label_columns(nodal_data: dict[str, _pd.DataFrame]) -> dict[str, _pd.DataFrame]:
    """Provides new columns with unique labels for "label", "to", "from", "connect"
    This function also adds a unique label for the buildings in "building_model_parameters".

    The DataFrames in the sheets are updated automatically as each df is a pointer to that sheet.
    """
    sheets = list(nodal_data.keys())
    building_sheet_name = ent.NodeKeysOptional.building_model_parameters

    if building_sheet_name in sheets:
        sheets.remove(building_sheet_name)
        df = nodal_data[building_sheet_name]
        df[ent.BuildingModelParameters.building_unique] = get_unique_buildings(df)

    sheets_with_no_need_for_unique_labels = [ent.NodeKeys.ice_storage, ent.NodeKeys.stratified_storage,
                                             ent.NodeKeys.profiles, ent.NodeKeys.links]
    # TODO: maybe remove time_step_data from sheets as well.

    [sheets.remove(x) for x in sheets_with_no_need_for_unique_labels if x in sheets]
    for sheet in sheets:
        df = nodal_data[sheet]
        df[ent.CommonLabels.label_unique] = get_unique_labels(
            df[[ent.CommonLabels.label, ent.CommonLabels.building]])

        if ent.CommonLabels.from_bus in df.columns:
            df[ent.CommonLabels.from_unique] = get_unique_buses(
                df[[ent.CommonLabels.from_bus, ent.CommonLabels.building]], ent.CommonLabels.from_bus)

        if ent.CommonLabels.to in df.columns:
            df[ent.CommonLabels.to_unique] = get_unique_buses(df[[ent.CommonLabels.to, ent.CommonLabels.building]],
                                                              ent.CommonLabels.to)

        if ent.CommonLabels.connect in df.columns:
            df[ent.CommonLabels.connect_unique] = get_unique_buses(
                df[[ent.CommonLabels.connect, ent.CommonLabels.building]], ent.CommonLabels.connect)

    return nodal_data


def get_unique_labels(df: _pd.DataFrame) -> list[str]:
    return [f"{row[ent.CommonLabels.label.value]}__B{str(row[ent.CommonLabels.building.value]).zfill(3)}" for _, row
            in df.iterrows()]


def get_unique_buses(df: _pd.DataFrame, buses_column: str) -> list[list[str]]:
    """Returns lists of strings for all cases, to simplify usage later."""
    buses = []
    for _, row in df.iterrows():
        row_buses = []
        for bus in row[buses_column].split(","):
            row_buses.append(f"{bus}__B{str(row[ent.CommonLabels.building.value]).zfill(3)}")
        buses.append(row_buses)
    return buses


def get_unique_buildings(df: _pd.DataFrame) -> list[str]:
    """Returns strings for both cases, to simplify usage later."""
    if ent.BuildingModelParameters.Circuit in df.columns:
        return [
            (f"Building_model__B{str(row[ent.BuildingModelParameters.Building_Number]).zfill(3)}"
             f"_C{str(row[ent.BuildingModelParameters.Circuit]).zfill(3)}")
            for _, row in df.iterrows()]

    return [f"Building_model__B{str(row[ent.BuildingModelParameters.Building_Number]).zfill(3)}"
            for _, row in df.iterrows()]


class ProfileAndOtherDataReader:
    @staticmethod
    def get_values_from_dataframe(df: _pd.DataFrame, identifier: str, identifier_column: str, desired_column: str,
                                  message: str,
                                  desired_instances: _tp.Tuple = (str, float, int, _np.int64, _np.float64),
                                  nan_allowed: bool = False,
                                  ) -> float | int | str:
        f""" Find any {identifier} entries in the {identifier_column} and return the 
             {desired_column} value of those rows."""
        row_indices = df[identifier_column] == identifier
        values = df.loc[row_indices, desired_column].iloc[0]

        try:
            # This makes this method less reusable.
            # It could be put into a wrapper function instead.
            # We currently pass a tuple for the assert for flexibility instead.
            assert isinstance(values, desired_instances)
        except AssertionError:
            message += f"\n Corrupt value in: \n{df.loc[row_indices]}"
            _log.error(message)
            raise ValueError(message)

        if not nan_allowed and isinstance(values, (float, _np.float64)) and _np.isnan(values):
            message += f"\n Value empty in: \n{df.loc[row_indices]}"
            _log.error(message)
            raise ValueError(message)

        return values

    def read_profiles_and_other_data(self, nodal_data: dict[str, _pd.DataFrame], file_or_folder_path: _pl.Path,
                                     num_buildings: int, cluster_size: dict[str, int] | None,
                                     time_index: _pd.DatetimeIndex,
                                     ) -> dict[str, _pd.DataFrame | dict[str, _pd.DataFrame]]:

        self.drop_inactive_rows(nodal_data)
        self.update_indices(nodal_data)

        # ====================================================
        # extract input data from CSVs
        # TODO: how would this change when getting data from an api connection?
        nodal_data = self.add_demand_profiles(nodal_data, cluster_size, num_buildings, time_index)
        nodal_data = self.add_electricity_impact(nodal_data, cluster_size, time_index)
        nodal_data = self.add_electricity_cost(nodal_data, cluster_size, time_index)
        nodal_data = self.add_weather_profiles(nodal_data, cluster_size, time_index)

        nodal_data = self.maybe_add_natural_gas(nodal_data, cluster_size, time_index)
        nodal_data = self.maybe_add_building_model_with_internal_gains(nodal_data, num_buildings, cluster_size,
                                                                       time_index)
        nodal_data = self.maybe_add_fixed_source_profiles(nodal_data, cluster_size, time_index)
        # ====================================================

        _log.info(f"Data from file {file_or_folder_path} imported.")

        return nodal_data

    @staticmethod
    def drop_inactive_rows(nodal_data):
        for key, df in nodal_data.items():
            if ent.BusesLabels.active not in df.columns:
                continue
            nodal_data[key] = df.where(df[ent.BusesLabels.active] == 1).dropna(how="all")
        return nodal_data

    @staticmethod
    def update_indices(nodal_data):
        """Stratified storages need the labels as indices."""
        # TODO: adjust this at the reading stage.  # pylint: disable=fixme
        nodal_data[ent.NodeKeys.stratified_storage.value].set_index(ent.StratifiedStorageLabels.label.value,
                                                                    inplace=True)
        if ent.NodeKeys.ice_storage.value in nodal_data:
            nodal_data[ent.NodeKeys.ice_storage.value].set_index(ent.IceStorageLabels.label.value, inplace=True)

    def add_weather_profiles(self, nodesData, clusterSize, time_index: _pd.DatetimeIndex):
        weatherDataPath = self.get_values_from_dataframe(
            df=nodesData[ent.NodeKeys.profiles],
            identifier_column=ent.ProfileLabels.name,
            identifier=ent.ProfileTypes.weather,
            desired_column=ent.ProfileLabels.path,
            message="Error in weather data file path",
        )

        if not _os.path.exists(weatherDataPath):
            _log.error("Error in weather data file path")
            raise FileNotFoundError(weatherDataPath)

        nodesData["weather_data"] = _pd.read_csv(weatherDataPath, delimiter=";")
        # add a timestamp column to the dataframe
        for index, row in nodesData['weather_data'].iterrows():
            time = f"{int(row['time.yy'])}.{int(row['time.mm']):02}.{int(row['time.dd']):02} {int(row['time.hh']
                                                                                                  ):02}:00:00"
            nodesData['weather_data'].at[index, 'timestamp'] = _dt.datetime.strptime(time, "%Y.%m.%d  %H:%M:%S")
            # set datetime index
        nodesData["weather_data"].timestamp = _pd.to_datetime(nodesData["weather_data"].timestamp,
                                                              format='%Y.%m.%d %H:%M:%S')
        nodesData["weather_data"].set_index("timestamp", inplace=True)
        if not clusterSize:
            # for data with typical years; we might have 2 years if summer of 1st yr and winter of 2nd yr is considered
            # we need to change index if len > 2 years
            if nodesData["weather_data"].index.year.unique().__len__() > 2:
                new_index = _pd.to_datetime({
                    'year': time_index.year[0],
                    'month': nodesData["weather_data"].index.month,
                    'day': nodesData["weather_data"].index.day,
                    'hour': nodesData["weather_data"].index.hour})
                nodesData["weather_data"].index = new_index
            nodesData["weather_data"] = self.clip_to_time_index(nodesData["weather_data"], time_index)

        if clusterSize:
            weatherData = _pd.concat([nodesData['weather_data'][
                                         nodesData['weather_data']['time.mm'] == int(d.split('-')[1])][
                                         nodesData['weather_data']['time.dd'] == int(d.split('-')[2])][
                                         ['gls', 'str.diffus', 'tre200h0', 'ground_temp']] for d in clusterSize.keys()])

            nodesData["weather_data"] = weatherData

        return nodesData

    def add_electricity_cost(self, nodesData, cluster_size, time_index: _pd.DatetimeIndex):
        electricityCost = self.get_values_from_dataframe(
            df=nodesData[ent.NodeKeys.commodity_sources],
            identifier_column=ent.CommonLabels.label,
            identifier=ent.CommoditySourceTypes.electricityResource,
            desired_column=ent.CommoditySourcesLabels.variable_costs,
            message="Error in electricity cost."
        )

        # int64 is not an instance of int!!
        # float64 is not an instance of float!!
        if isinstance(electricityCost, (float, int, _np.float64, _np.int64)):
            # for constant cost
            electricityCostValue = electricityCost
            _log.info("Constant value for electricity cost")
            nodesData["electricity_cost"] = _pd.DataFrame()
            nodesData["electricity_cost"]["cost"] = (nodesData["demandProfiles"][1].shape[0]) * [
                electricityCostValue]
            nodesData["electricity_cost"].index = nodesData["demandProfiles"][1].index
        elif not _os.path.exists(electricityCost):
            _log.error("Error in electricity cost file path")
            raise FileNotFoundError(electricityCost)
        else:
            nodesData["electricity_cost"] = _pd.read_csv(electricityCost, delimiter=";")
            # set datetime index
            nodesData["electricity_cost"].set_index("timestamp", inplace=True)
            nodesData["electricity_cost"].index = _pd.to_datetime(nodesData["electricity_cost"].index,
                                                                  format='%d.%m.%Y %H:%M')
            nodesData["electricity_cost"] = self.clip_to_time_index(nodesData["electricity_cost"], time_index)

        if cluster_size:
            electricityCost = self.cluster_and_multiply_desired_column(nodesData["electricity_cost"], cluster_size)
            nodesData["electricity_cost"] = electricityCost

        return nodesData

    def add_electricity_impact(self, nodesData, cluster_size, time_index: _pd.DatetimeIndex):
        electricityImpact = self.get_values_from_dataframe(
            df=nodesData[ent.NodeKeys.commodity_sources],
            identifier_column=ent.CommoditySourcesLabels.label,
            identifier=ent.CommoditySourceTypes.electricityResource,
            desired_column=ent.CommoditySourcesLabels.CO2_impact,
            message="Error in electricity impact.",
        )

        if isinstance(electricityImpact, (float, int, _np.float64, _np.int64)):
            # for constant impact
            electricityImpactValue = electricityImpact
            _log.info("Constant value for electricity impact")
            nodesData["electricity_impact"] = _pd.DataFrame()
            nodesData["electricity_impact"]["impact"] = (nodesData["demandProfiles"][1].shape[0]) * [
                electricityImpactValue]
            nodesData["electricity_impact"].index = nodesData["demandProfiles"][1].index
        elif not _os.path.exists(electricityImpact):
            _log.error("Error in electricity impact file path")
            raise FileNotFoundError("Error in electricity impact file path")
        else:
            nodesData["electricity_impact"] = _pd.read_csv(electricityImpact, delimiter=";")
            # set datetime index
            nodesData["electricity_impact"].set_index("timestamp", inplace=True)
            nodesData["electricity_impact"].index = _pd.to_datetime(nodesData["electricity_impact"].index,
                                                                    format='%d.%m.%Y %H:%M')
            nodesData["electricity_impact"] = self.clip_to_time_index(nodesData["electricity_impact"], time_index)

        if cluster_size:
            electricityImpact = self.cluster_and_multiply_desired_column(nodesData["electricity_impact"], cluster_size)
            nodesData["electricity_impact"] = electricityImpact

        return nodesData

    @staticmethod
    def cluster_desired_column(df: _pd.DataFrame, clusterSize: dict[str, int]) -> _pd.DataFrame:
        return _pd.concat([df.loc[d] for d in clusterSize.keys()])

    @staticmethod
    def cluster_and_multiply_desired_column(df: _pd.DataFrame,  cluster_size: dict[str, int]) -> _pd.DataFrame:
        return _pd.concat([df.loc[d] * v for d, v in cluster_size.items()])

    def add_demand_profiles(self, nodesData, clusterSize, numBuildings, time_index: _pd.DatetimeIndex):
        # demandProfilesPath should contain csv file(s) (one for each building's profiles)
        demandProfilesPath = self.get_values_from_dataframe(
            df=nodesData[ent.NodeKeys.profiles],
            identifier_column=ent.ProfileLabels.name,
            identifier=ent.ProfileTypes.demand,
            desired_column=ent.ProfileLabels.path,
            message="Error in the demand profiles path."
        )

        if not _os.path.exists(demandProfilesPath) or not _os.listdir(demandProfilesPath):
            _log.error(f"Error in the demand profiles path: The folder is either empty or does not exist: "
                       f"{demandProfilesPath}")
            raise FileNotFoundError(f"Error in the demand profiles path: The folder is either empty or does not exist: "
                                    f"{demandProfilesPath}")

        demandProfiles = {}  # dictionary of dataframes for each building's demand profiles
        i = 0
        for filename in _os.listdir(demandProfilesPath):
            i += 1  # Building number
            if i > numBuildings:
                _log.warning("Demand profiles folder has more files than the number of buildings specified")
                break
            demandProfiles.update({i: _pd.read_csv(_os.path.join(demandProfilesPath, filename), delimiter=";")})

        nodesData["demandProfiles"] = demandProfiles

        # set datetime index
        for i in range(numBuildings):
            nodesData["demandProfiles"][i + 1].timestamp = _pd.to_datetime(nodesData["demandProfiles"][i + 1].timestamp,
                                                                           format='%Y-%m-%d %H:%M:%S')
            nodesData["demandProfiles"][i + 1].set_index("timestamp", inplace=True)
            if not clusterSize:
                nodesData["demandProfiles"][i + 1] = self.clip_to_time_index(nodesData["demandProfiles"][i + 1],
                                                                             time_index)
        if clusterSize:
            demandProfiles = {}
            for i in range(1, numBuildings + 1):
                demandProfiles[i] = self.cluster_desired_column(nodesData["demandProfiles"][i], clusterSize)
            nodesData["demandProfiles"] = demandProfiles

        return nodesData

    def maybe_add_natural_gas(self, nodesData, cluster_size, time_index: _pd.DatetimeIndex):
        if "naturalGasResource" not in nodesData["commodity_sources"]["label"].values:
            return nodesData

        # Making this a wrapper function simplifies testing.
        nodesData = self.add_natural_gas_impact(nodesData, cluster_size, time_index)
        nodesData = self.add_natural_gas_costs(nodesData, cluster_size, time_index)

        return nodesData

    def maybe_add_fixed_source_profiles(self, nodesData, cluster_size, time_index: _pd.DatetimeIndex):
        if not "fixed" in nodesData["commodity_sources"].columns:
            return nodesData
        if not (nodesData["commodity_sources"]["fixed"].eq(1)).any():
            return nodesData

        fixed_source_profiles_path = self.get_values_from_dataframe(
            df=nodesData[ent.NodeKeys.profiles],
            identifier_column=ent.ProfileLabels.name,
            identifier=ent.ProfileTypes.fixed_sources,
            desired_column=ent.ProfileLabels.path,
            message="Error in the fixed source profiles path."
        )

        if not _os.path.exists(fixed_source_profiles_path) or not _os.listdir(fixed_source_profiles_path):
            _log.error(f"Error in the fixed sources profile path: The folder is either empty or does not exist: "
                       f"{fixed_source_profiles_path}")
            raise FileNotFoundError(f"Error in the fixed sources profile path: The folder is either empty or does not exist: "
                                    f"{fixed_source_profiles_path}")

        fixed_sources_data = {}
        for filename in _os.listdir(fixed_source_profiles_path):
            fixed_sources_data.update({filename.split(".csv")[0]: _pd.read_csv(_os.path.join(fixed_source_profiles_path, filename), delimiter=";")})

        nodesData["fixed_sources"] = fixed_sources_data

        # set datetime index
        for i in nodesData["fixed_sources"].keys():
            nodesData["fixed_sources"][i].timestamp = _pd.to_datetime(nodesData["fixed_sources"][i].timestamp,
                                                                           format='%Y-%m-%d %H:%M:%S')
            nodesData["fixed_sources"][i].set_index("timestamp", inplace=True)
            if not cluster_size:
                nodesData["fixed_sources"][i] = self.clip_to_time_index(nodesData["fixed_sources"][i],
                                                                             time_index)
        if cluster_size:
            fixed_sources_data = {}
            for i in nodesData["fixed_sources"].keys():
                fixed_sources_data[i] = self.cluster_desired_column(nodesData["fixed_sources"][i], cluster_size)
            nodesData["fixed_sources"] = fixed_sources_data

        return nodesData

    def add_natural_gas_costs(self, nodesData, cluster_size, time_index):
        natGasCost = self.get_values_from_dataframe(
            df=nodesData[ent.NodeKeys.commodity_sources],
            identifier_column=ent.CommonLabels.label,
            identifier=ent.CommoditySourceTypes.naturalGasResource,
            desired_column=ent.CommoditySourcesLabels.variable_costs,
            message="Error in natural gas cost.",
        )
        if isinstance(natGasCost, (int, float, _np.float64, _np.int64)):
            # for constant cost
            natGasCostValue = natGasCost
            _log.info("Constant value for natural gas cost")
            nodesData["natGas_cost"] = _pd.DataFrame()
            nodesData["natGas_cost"]["cost"] = (nodesData["demandProfiles"][1].shape[0]) * [natGasCostValue]
            nodesData["natGas_cost"].index = nodesData["demandProfiles"][1].index
        elif not _os.path.exists(natGasCost):
            _log.error("Error in natural gas cost file path")
            raise FileNotFoundError(f"Error in natural gas cost file path: {natGasCost}")

        else:
            nodesData["natGas_cost"] = _pd.read_csv(natGasCost, delimiter=";")
            # set datetime index
            nodesData["natGas_cost"].set_index("timestamp", inplace=True)
            nodesData["natGas_cost"].index = _pd.to_datetime(nodesData["natGas_cost"].index, format='%d.%m.%Y %H:%M')
            nodesData["natGas_cost"] = self.clip_to_time_index(nodesData["natGas_cost"], time_index)
        if cluster_size:
            natGasCost = self.cluster_and_multiply_desired_column(nodesData["natGas_cost"], cluster_size)
            nodesData["natGas_cost"] = natGasCost

        return nodesData

    def add_natural_gas_impact(self, nodesData, cluster_size, time_index):
        natGasImpact = self.get_values_from_dataframe(
            df=nodesData[ent.NodeKeys.commodity_sources],
            identifier_column=ent.CommonLabels.label,
            identifier=ent.CommoditySourceTypes.naturalGasResource,
            desired_column=ent.CommoditySourcesLabels.CO2_impact,
            message="Error in natural gas impact.",
        )
        if isinstance(natGasImpact, (int, float, _np.float64, _np.int64)) or (
                natGasImpact.split('.')[0].replace('-', '').isdigit() and
                natGasImpact.split('.')[1].replace('-', '').isdigit()
        ):
            # for constant impact
            natGasImpactValue = float(natGasImpact)
            _log.info("Constant value for natural gas impact")
            nodesData["natGas_impact"] = _pd.DataFrame()
            nodesData["natGas_impact"]["impact"] = (nodesData["demandProfiles"][1].shape[0]) * [
                natGasImpactValue]
            nodesData["natGas_impact"].index = nodesData["demandProfiles"][1].index
        elif not _os.path.exists(natGasImpact):
            _log.error("Error in natural gas impact file path")
            raise FileNotFoundError(f"Error in natural gas impact file path: {natGasImpact}")
        else:
            nodesData["natGas_impact"] = _pd.read_csv(natGasImpact, delimiter=";")
            # set datetime index
            nodesData["natGas_impact"].set_index("timestamp", inplace=True)
            nodesData["natGas_impact"].index = _pd.to_datetime(nodesData["natGas_impact"].index,
                                                               format='%d.%m.%Y %H:%M')
            nodesData["natGas_impact"] = self.clip_to_time_index(nodesData["natGas_impact"], time_index)
        if cluster_size:
            natGasImpact = self.cluster_and_multiply_desired_column(nodesData["natGas_impact"], cluster_size)
            nodesData["natGas_impact"] = natGasImpact

        return nodesData

    def maybe_add_building_model_with_internal_gains(self, nodesData, numBuildings, clusterSize,
                                                     time_index: _pd.DatetimeIndex):
        nodesData["building_model"] = {}
        if not nodesData['demand']['building model'].notna().any() or not (
                nodesData['demand']['building model'] == 'Yes').any():
            _log.info("Building model either not selected or invalid string value entered")
        else:
            nodesData, internalGains = self.add_internal_gains(nodesData, clusterSize, time_index)

            nodesData = self.add_building_models(nodesData, numBuildings, internalGains)

        return nodesData

    def add_internal_gains(self, nodesData, clusterSize, time_index):
        internalGainsPath = self.get_values_from_dataframe(
            df=nodesData[ent.NodeKeys.profiles],
            identifier_column=ent.ProfileLabels.name,
            identifier=ent.ProfileTypes.internal_gains,
            desired_column=ent.ProfileLabels.path,
            message="Error in internal gains file path for the building model."
        )
        if not _os.path.exists(internalGainsPath):
            _log.error("Error in internal gains file path for the building model.")
            raise FileNotFoundError(f"Error in internal gains file path for the building model{internalGainsPath}")
        internalGains = _pd.read_csv(internalGainsPath, delimiter=';')
        internalGains.timestamp = _pd.to_datetime(internalGains.timestamp, format='%d.%m.%Y %H:%M')
        internalGains.set_index("timestamp", inplace=True)
        if not clusterSize:
            internalGains = self.clip_to_time_index(internalGains, time_index)
        return nodesData, internalGains

    def add_building_models(self, nodesData, numBuildings, internalGains):
        b_model_params_Path = self.get_values_from_dataframe(
            df=nodesData[ent.NodeKeys.profiles],
            identifier_column=ent.ProfileLabels.name,
            identifier=ent.ProfileTypes.building_model_params,
            desired_column=ent.ProfileLabels.path,
            message="Error in building model parameters file path."
        )
        if not _os.path.exists(b_model_params_Path):
            _log.error("Error in building model parameters file path.")
            raise FileNotFoundError(f"Error in building model parameters file path: {b_model_params_Path}")

        bmParamers = _pd.read_csv(b_model_params_Path, delimiter=';')
        for i in range(numBuildings):
            nodesData["building_model"][i + 1] = {}
            nodesData["building_model"][i + 1]["timeseries"] = _pd.DataFrame()
            nodesData["building_model"][i + 1]["timeseries"]["tAmb"] = _np.array(
                nodesData["weather_data"]["tre200h0"])
            nodesData["building_model"][i + 1]["timeseries"]["IrrH"] = _np.array(
                nodesData["weather_data"]["gls"]) / 1000  # conversion from W/m2 to kW/m2
            nodesData["building_model"][i + 1]["timeseries"][f"Qocc"] = internalGains[f'Total (kW) {i + 1}'].values

            if "tIndoorDay" in bmParamers.columns:
                tIndoorDay = float(bmParamers[bmParamers["Building Number"] == (i + 1)]['tIndoorDay'].iloc[0])
                tIndoorNight = float(bmParamers[bmParamers["Building Number"] == (i + 1)]['tIndoorNight'].iloc[0])
                tIndoorSet = [tIndoorNight, tIndoorNight, tIndoorNight, tIndoorNight, tIndoorNight, tIndoorNight,
                              tIndoorNight, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay,
                              tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay,
                              tIndoorNight, tIndoorNight, tIndoorNight, tIndoorNight] * 365
                nodesData["building_model"][i + 1]["timeseries"]["tIndoorSet"] = _pd.DataFrame(tIndoorSet).values
            paramList = ['gAreaWindows', 'rDistribution', 'cDistribution', 'rWall', 'cWall', 'rIndoor',
                         'cIndoor',
                         'qDistributionMin', 'qDistributionMax', 'tIndoorMin', 'tIndoorMax', 'tIndoorInit',
                         'tWallInit', 'tDistributionInit']

            for param in paramList:
                nodesData["building_model"][i + 1][param] = float(
                    bmParamers[bmParamers["Building Number"] == (i + 1)][param].iloc[0])

        return nodesData

    @staticmethod
    def clip_to_time_index(df: _pd.DataFrame, time_index: _pd.DatetimeIndex):
        return df[time_index[0]:time_index[-1]]



import json
import os as _os
import pathlib as _pl
import unittest as _ut

import pandas as _pd
import pytest as _pt

import optihood as _oh
import optihood.IO.readers as _re
import optihood.Visualizer.convert_scenario as _cs
import optihood.Visualizer.visualizer_app as _va
import optihood.energy_network as _en

cwd = _os.getcwd()
packageDir = _pl.Path(_oh.__file__).resolve().parent
_input_data_dir = packageDir / ".." / "data" / "excels" / "basic_example"
_examples_dir = packageDir / ".." / "data" / "examples"
_input_data_path = _input_data_dir / "scenario.xls"
filter_results = False   #False
include_links = True
building_no = 1

# Helper functions
def get_building_no(node):
    """Extract building number from node or return 0 if not present."""
    return int(node.split("_B")[-1]) if "_B" in node else 0

def is_link(node):
    """Check if the node ends with 'link' (case-insensitive)."""
    return "link" in node.lower()


@_pt.mark.manual
class TestVisualizeEnergyNetwork(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_visualize_from_energy_network(self):
        # Given
        time_period = _pd.date_range("2018-01-01 00:00:00", "2018-01-31 23:00:00", freq="60min")

        _os.chdir(_examples_dir)
        network = _en.EnergyNetworkClass(time_period)
        data = _pd.ExcelFile(str(_input_data_path))
        initial_nodal_data = network.get_nodal_data_from_Excel(data)
        data.close()
        _os.chdir(cwd)

        # When
        converters = _cs.get_converters(initial_nodal_data, nr_of_buildings=4)
        graphData = _cs.get_graph_data(converters)
        if filter_results:
            graphData.nodes = [n for n in graphData.nodes if
                               ('building' in n['data'] and n['data']['building'] == building_no)
                               or (include_links and n['classes'] == "link")]
            graphData.edges = [e for e in graphData.edges if
                               (get_building_no(e['data']['source']) == get_building_no(e['data']['target']) == building_no)
                               or (include_links and (is_link(e['data']['source']) or is_link(e['data']['target'])) and
                                   (
                                           get_building_no(e['data']['target']) == building_no or
                                           get_building_no(e['data']['source']) == building_no
                                   ))]
        _va.run_cytoscape_visualizer(graphData=graphData)

        self.assertEqual(True, False)  # add assertion here


@_pt.mark.manual
class TestVisualizeEnergyNetwork_from_csv(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_visualize_from_energy_network(self):
        # Given
        _input_data_dir = packageDir / ".." / "data" / "CSVs" / "basic_example_CSVs"
        _os.chdir(_examples_dir)
        csvReader = _re.CsvScenarioReader(_input_data_dir)
        initial_nodal_data = csvReader.read_scenario()
        _os.chdir(cwd)

        # When
        converters = _cs.get_converters(initial_nodal_data, nr_of_buildings=4)
        graphData = _cs.get_graph_data(converters)
        if filter_results:
            graphData.nodes = [n for n in graphData.nodes if
                               ('building' in n['data'] and n['data']['building'] == building_no)
                               or (include_links and n['classes'] == "link")]
            graphData.edges = [e for e in graphData.edges if
                               (get_building_no(e['data']['source']) == get_building_no(e['data']['target']) == building_no)
                               or (include_links and (is_link(e['data']['source']) or is_link(e['data']['target'])) and
                                   (
                                           get_building_no(e['data']['target']) == building_no or
                                           get_building_no(e['data']['source']) == building_no
                                   ))]
        _va.run_cytoscape_visualizer(graphData=graphData)

        self.assertEqual(True, False)  # add assertion here


class TestScenarioTillGraphData(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_example_scenario(self):
        input_data_path = _input_data_dir / "scenario.xls"

         # Given
        time_period = _pd.date_range("2018-01-01 00:00:00", "2018-01-31 23:00:00", freq="60min")

        _os.chdir(_examples_dir)
        network = _en.EnergyNetworkClass(time_period)
        data = _pd.ExcelFile(str(input_data_path))
        initial_nodal_data = network.get_nodal_data_from_Excel(data)
        data.close()
        _os.chdir(cwd)

        # When
        converters = _cs.get_converters(initial_nodal_data, nr_of_buildings=4)
        graphData = _cs.get_graph_data(converters)

        with open(_pl.Path(__file__).parent / 'expected_files' / 'test_example_scenario_nodes.json') as f:
            nodes_expected = json.load(f)

        assert len(graphData.nodes) == len(nodes_expected)
        self.assertListEqual(graphData.nodes, nodes_expected)

        with open(_pl.Path(__file__).parent / 'expected_files' / 'test_example_scenario_edges.json') as f:
            edges_expected = json.load(f)
        assert len(graphData.edges) == len(edges_expected)
        self.assertListEqual(graphData.edges, edges_expected)


if __name__ == '__main__':
    _ut.main()



from pyomo.core.base.block import ScalarBlock
from pyomo.environ import BuildAction
from pyomo.environ import Constraint
from optihood._helpers import *
from oemof.solph import components as solph_components
from oemof.solph._plumbing import sequence as solph_sequence

class Link(solph_components.Transformer):
    r"""
    A transformer to model links between different buildings
    The constraint defined in a new constraint block equates the sum of all the input flows
    to the sum of all the output flows
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def constraint_group(self):
        return LinkBlock

class LinkBlock(ScalarBlock):
    r"""Block for the linear relation of nodes of type Link"""
    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """Creates the two linear constraints for nodes of type Link
        """
        if group is None:
            return None

        m = self.parent_block()
        in_flows = {n: [i for i in n.inputs.keys()] for n in group}
        out_flows = {n: [o for o in n.outputs.keys()] for n in group}
        efficiency = [n.conversion_factors[next(iter(n.outputs.keys()))] for n in group][0][0]   # conversion factors of all the output flows of all the groups will be equal, therefore first value is chosen

        def _input_output_relation(block):
            """Constraint defining the relation between input and outputs."""
            self.input_output_relation = Constraint(group, m.TIMESTEPS, noruleinit=True)

            for t in m.TIMESTEPS:
                for g in group:
                    lhs = sum(m.flow[i, g, t] for i in in_flows[g]) * efficiency
                    rhs = sum(m.flow[g, o, t] for o in out_flows[g])
                    block.input_output_relation.add((g, t), (lhs == rhs))

        self.input_output_relation_build = BuildAction(rule=_input_output_relation)

class LinkTemperatureDemand(solph_components.Transformer):
    r"""
        A transformer to model links between different different temperature levels and demand sink component
        The constraint defined in a new constraint block equates the weighted sum of all the input flows
        to the output flow.
        """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def constraint_group(self):
        return LinkTemperatureDemandBlock


class LinkTemperatureDemandBlock(ScalarBlock):
    r"""Block for the linear relation of nodes of type LinkTemperatureDemand"""
    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """Creates the two linear constraints for nodes of type Link
        """
        if group is None:
            return None

        m = self.parent_block()
        in_flows = {n: [i for i in n.inputs.keys()] for n in group}
        out_flows = {n: [o for o in n.outputs.keys()] for n in group}

        def _input_output_relation(block):
            """Constraint defining the relation between input and outputs."""
            self.input_output_relation = Constraint(group, m.TIMESTEPS, noruleinit=True)

            for t in m.TIMESTEPS:
                for g in group:
                    lhs = sum(m.flow[i, g, t]*g.conversion_factors[i][t] for i in in_flows[g])
                    rhs = sum(m.flow[g, o, t] for o in out_flows[g])
                    block.input_output_relation.add((g, t), (lhs == rhs))

        self.input_output_relation_build = BuildAction(rule=_input_output_relation)

class LinkStorageDummyInput(solph_components.Transformer):
    r"""
        A transformer to model links between output from different temperature levels and inputs in multi-temperature
        level thermal storages. This transformer takes two input flows, one coming from energy converter and the other
        from a storage volume.
        The constraint defined in a new constraint block equates the two input flows to ensure volume from lower
        temperature to higher temperature within the storage. Moreover, the output flow would be equal to the input flow.
        input_flow_1 == input_flow_2 == output_flow
        """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def constraint_group(self):
        return LinkStorageDummyInputBlock


class LinkStorageDummyInputBlock(ScalarBlock):
    r"""Block for the linear relation of nodes of type LinkStorageDummyInput"""
    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """Creates the two linear constraints for nodes of type Link
        """
        if group is None:
            return None

        m = self.parent_block()
        in_flows = {n: [i for i in n.inputs.keys()] for n in group}
        out_flows = {n: [o for o in n.outputs.keys()] for n in group}

        def _input_relation(block):
            """Constraint defining the relation between the two inputs of the component.
            First input is heat input for the storage, second input is the heat input from
            the preceding (lower) temperature level storage"""
            self.input_relation = Constraint(group, m.TIMESTEPS, noruleinit=True)

            for t in m.TIMESTEPS:
                for g in group:
                    lhs = m.flow[in_flows[g][0], g, t]*g.conversion_factors[in_flows[g][0]][t]
                    rhs = m.flow[in_flows[g][1], g, t]*g.conversion_factors[in_flows[g][1]][t]
                    block.input_relation.add((g, t), (lhs == rhs))

        self.input_relation_build = BuildAction(rule=_input_relation)

        def _input_output_relation(block):
            """Constraint defining the relation between two inputs and one output of the component."""
            self.input_output_relation = Constraint(group, m.TIMESTEPS, noruleinit=True)

            for t in m.TIMESTEPS:
                for g in group:
                    lhs = m.flow[in_flows[g][0], g, t]*g.conversion_factors[in_flows[g][0]][t]
                    rhs = m.flow[g, out_flows[g][0], t]
                    block.input_output_relation.add((g, t), (lhs == rhs))

        self.input_output_relation_build = BuildAction(rule=_input_output_relation)


import numbers

from oemof.network import network
from pyomo.core.base.block import ScalarBlock
from pyomo.environ import Binary
from pyomo.environ import Constraint
from pyomo.environ import Expression
from pyomo.environ import NonNegativeReals, PositiveReals
from pyomo.environ import Set
from pyomo.environ import Var

from oemof.solph._helpers import check_node_object_for_missing_attribute
from oemof.solph._options import Investment
from oemof.solph._plumbing import sequence as solph_sequence


class GenericStoragePit(network.Node):
    r"""
    Component `GenericStoragePit`, derived from `GenericStorage`from oemof.solph

    The energy balance equations represent those of a pit storage.

    The GenericStoragePit is designed for one input and one output.

    Parameters
    ----------
    height_multiplication_factor : numeric
        factor for the calculation of height from nominal storage capacity
    temp_h : numeric
        temperature of the hot layer in the storage [°C]
    temp_c : numeric
        return temperature or temperature of the cold layer in the storage [°C]
    rho : numeric
        Density of the fluid in kg/L
    c : numeric
        Specific heat capacity of the fluid in kJ/(kg K)
    nominal_storage_capacity : numeric, :math:`E_{nom}`
        Absolute nominal capacity of the storage
    invest_relation_input_capacity : numeric or None, :math:`r_{cap,in}`
        Ratio between the investment variable of the input Flow and the
        investment variable of the storage:
        :math:`\dot{E}_{in,invest} = E_{invest} \cdot r_{cap,in}`
    invest_relation_output_capacity : numeric or None, :math:`r_{cap,out}`
        Ratio between the investment variable of the output Flow and the
        investment variable of the storage:
        :math:`\dot{E}_{out,invest} = E_{invest} \cdot r_{cap,out}`
    invest_relation_input_output : numeric or None, :math:`r_{in,out}`
        Ratio between the investment variable of the output Flow and the
        investment variable of the input flow. This ratio used to fix the
        flow investments to each other.
        Values < 1 set the input flow lower than the output and > 1 will
        set the input flow higher than the output flow. If None no relation
        will be set:
        :math:`\dot{E}_{in,invest} = \dot{E}_{out,invest} \cdot r_{in,out}`
    initial_storage_level : numeric, :math:`c(-1)`
        The relative storage content in the timestep before the first
        time step of optimization (between 0 and 1).
    balanced : boolean
        Couple storage level of first and last time step.
        (Total inflow and total outflow are balanced.)
    loss_rate : numeric (iterable or scalar)
        The relative loss of the storage content per hour.
    fixed_losses_relative : numeric (iterable or scalar), :math:`\gamma(t)`
        Losses per hour that are independent of the storage content but
        proportional to nominal storage capacity.
    fixed_losses_absolute : numeric (iterable or scalar), :math:`\delta(t)`
        Losses per hour that are independent of storage content and independent
        of nominal storage capacity.
    inflow_conversion_factor : numeric (iterable or scalar), :math:`\eta_i(t)`
        The relative conversion factor, i.e. efficiency associated with the
        inflow of the storage.
    outflow_conversion_factor : numeric (iterable or scalar), :math:`\eta_o(t)`
        see: inflow_conversion_factor
    min_storage_level : numeric (iterable or scalar), :math:`c_{min}(t)`
        The normed minimum storage content as fraction of the
        nominal storage capacity (between 0 and 1).
        To set different values in every time step use a sequence.
    max_storage_level : numeric (iterable or scalar), :math:`c_{max}(t)`
        see: min_storage_level
    investment : :class:`oemof.solph.options.Investment` object
        Object indicating if a nominal_value of the flow is determined by
        the optimization problem. Note: This will refer all attributes to an
        investment variable instead of to the nominal_storage_capacity. The
        nominal_storage_capacity should not be set (or set to None) if an
        investment object is used.
    """

    def __init__(
        self,
        label=None,
        inputs=None,
        outputs=None,
        nominal_storage_capacity=None,
        height_multiplication_factor=None,
        temp_h=None,
        temp_c=None,
        rho=None,
        c=None,
        initial_storage_level=None,
        investment=None,
        invest_relation_input_output=None,
        invest_relation_input_capacity=None,
        invest_relation_output_capacity=None,
        min_storage_level=0,
        max_storage_level=1,
        balanced=True,
        loss_rate=0,
        fixed_losses_relative=0,
        fixed_losses_absolute=0,
        inflow_conversion_factor=1,
        outflow_conversion_factor=1,
        custom_attributes=None,
    ):
        if inputs is None:
            inputs = {}
        if outputs is None:
            outputs = {}
        if custom_attributes is None:
            custom_attributes = {}
        super().__init__(
            label=label,
            inputs=inputs,
            outputs=outputs,
            **custom_attributes,
        )

        self.nominal_storage_capacity = None
        self.height = None
        self.investment = None
        self._invest_group = False
        if isinstance(nominal_storage_capacity, numbers.Real):
            self.nominal_storage_capacity = nominal_storage_capacity
            self.height = (self.nominal_storage_capacity * 3600 * 1e3 / height_multiplication_factor) ** (1 / 3)
        elif isinstance(nominal_storage_capacity, Investment):
            self.investment = nominal_storage_capacity
            self._invest_group = True
        self.initial_storage_level = initial_storage_level
        self.height_multiplication_factor = height_multiplication_factor
        self.temp_h = temp_h
        self.temp_c = temp_c
        self.rho = rho
        self.c = c
        self.balanced = balanced
        self.loss_rate = solph_sequence(loss_rate)
        self.fixed_losses_relative = solph_sequence(fixed_losses_relative)
        self.fixed_losses_absolute = solph_sequence(fixed_losses_absolute)
        self.inflow_conversion_factor = solph_sequence(
            inflow_conversion_factor
        )
        self.outflow_conversion_factor = solph_sequence(
            outflow_conversion_factor
        )
        self.max_storage_level = solph_sequence(max_storage_level)
        self.min_storage_level = solph_sequence(min_storage_level)
        self.investment = investment
        self.invest_relation_input_output = invest_relation_input_output
        self.invest_relation_input_capacity = invest_relation_input_capacity
        self.invest_relation_output_capacity = invest_relation_output_capacity
        self._invest_group = isinstance(self.investment, Investment)

        # Check number of flows.
        self._check_number_of_flows()
        # Check for infeasible parameter combinations
        self._check_infeasible_parameter_combinations()

        # Check attributes for the investment mode.
        if self._invest_group is True:
            self._check_invest_attributes()

    def _set_flows(self):
        for flow in self.inputs.values():
            if (
                self.invest_relation_input_capacity is not None
                and not isinstance(flow.investment, Investment)
            ):
                flow.investment = Investment()
        for flow in self.outputs.values():
            if (
                self.invest_relation_output_capacity is not None
                and not isinstance(flow.investment, Investment)
            ):
                flow.investment = Investment()

    def _check_invest_attributes(self):
        if self.investment and self.nominal_storage_capacity is not None:
            e1 = (
                "If an investment object is defined the invest variable "
                "replaces the nominal_storage_capacity.\n Therefore the "
                "nominal_storage_capacity should be 'None'.\n"
            )
            raise AttributeError(e1)
        if (
            self.invest_relation_input_output is not None
            and self.invest_relation_output_capacity is not None
            and self.invest_relation_input_capacity is not None
        ):
            e2 = (
                "Overdetermined. Three investment object will be coupled"
                "with three constraints. Set one invest relation to 'None'."
            )
            raise AttributeError(e2)
        if (
            self.investment
            and sum(solph_sequence(self.fixed_losses_absolute)) != 0
            and self.investment.existing == 0
            and self.investment.minimum == 0
        ):
            e3 = (
                "With fixed_losses_absolute > 0, either investment.existing "
                "or investment.minimum has to be non-zero."
            )
            raise AttributeError(e3)

        self._set_flows()

    def _check_number_of_flows(self):
        msg = "Only one {0} flow allowed in the GenericStorage {1}."
        check_node_object_for_missing_attribute(self, "inputs")
        check_node_object_for_missing_attribute(self, "outputs")
        if len(self.inputs) > 1:
            raise AttributeError(msg.format("input", self.label))
        if len(self.outputs) > 1:
            raise AttributeError(msg.format("output", self.label))

    def _check_infeasible_parameter_combinations(self):
        """Checks for infeasible parameter combinations and raises error"""
        msg = (
            "initial_storage_level must be greater or equal to "
            "min_storage_level and smaller or equal to "
            "max_storage_level."
        )
        if self.initial_storage_level is not None:
            if (
                self.initial_storage_level < self.min_storage_level[0]
                or self.initial_storage_level > self.max_storage_level[0]
            ):
                raise ValueError(msg)

    def constraint_group(self):
        if self._invest_group is True:
            return GenericInvestmentStorageBlockPit
        else:
            return GenericStorageBlockPit


class GenericStorageBlockPit(ScalarBlock):
    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """
        Parameters
        ----------
        group : list
            List containing storage objects.
            e.g. groups=[storage1, storage2,..]
        """
        m = self.parent_block()

        if group is None:
            return None

        i = {n: [i for i in n.inputs][0] for n in group}
        o = {n: [o for o in n.outputs][0] for n in group}

        #  ************* SETS *********************************

        self.STORAGES = Set(initialize=[n for n in group])

        self.STORAGES_BALANCED = Set(
            initialize=[n for n in group if n.balanced is True]
        )

        self.STORAGES_INITITAL_LEVEL = Set(
            initialize=[
                n for n in group if n.initial_storage_level is not None
            ]
        )

        self.STORAGES_WITH_INVEST_FLOW_REL = Set(
            initialize=[
                n for n in group if n.invest_relation_input_output is not None
            ]
        )

        #  ************* VARIABLES *****************************

        def _storage_content_bound_rule(block, n, t):
            """
            Rule definition for bounds of storage_content variable of
            storage n in timestep t.
            """
            bounds = (
                n.nominal_storage_capacity * n.min_storage_level[t],
                n.nominal_storage_capacity * n.max_storage_level[t],
            )
            return bounds

        self.storage_content = Var(
            self.STORAGES, m.TIMEPOINTS, bounds=_storage_content_bound_rule
        )
        self.height = Var(self.STORAGES)

        # set the initial storage content
        for n in group:
            if n.initial_storage_level is not None:
                self.storage_content[n, 0] = (
                    n.initial_storage_level * n.nominal_storage_capacity
                )
                self.storage_content[n, 0].fix()

        #  ************* Constraints ***************************

        def _storage_balance_rule(block, n, t):
            """
            Rule definition for the storage balance of every storage n and
            every timestep.
            """
            expr = 0
            expr += block.storage_content[n, t + 1]
            expr += (
                -block.storage_content[n, t]
                * (1 - n.loss_rate[t] / n.height) ** m.timeincrement[t]
            )
            expr += (
                n.fixed_losses_relative[t] / n.height
                * n.nominal_storage_capacity
                * m.timeincrement[t]
            )
            expr += n.fixed_losses_absolute[t] * m.timeincrement[t]
            expr += (
                -m.flow[i[n], n, t] * n.inflow_conversion_factor[t]
            ) * m.timeincrement[t]
            expr += (
                m.flow[n, o[n], t] / n.outflow_conversion_factor[t]
            ) * m.timeincrement[t]
            return expr == 0

        self.balance = Constraint(
            self.STORAGES, m.TIMESTEPS, rule=_storage_balance_rule
        )

        def _balanced_storage_rule(block, n):
            """
            Storage content of last time step == initial storage content
            if balanced.
            """
            return (
                block.storage_content[n, m.TIMEPOINTS.at(-1)]
                == block.storage_content[n, m.TIMEPOINTS.at(1)]
            )

        self.balanced_cstr = Constraint(
            self.STORAGES_BALANCED, rule=_balanced_storage_rule
        )

        def _power_coupled(block, n):
            """
            Rule definition for constraint to connect the input power
            and output power
            """
            expr = (
                m.InvestmentFlowBlock.invest[n, o[n]]
                + m.flows[n, o[n]].investment.existing
            ) * n.invest_relation_input_output == (
                m.InvestmentFlowBlock.invest[i[n], n]
                + m.flows[i[n], n].investment.existing
            )
            return expr

        self.power_coupled = Constraint(
            self.STORAGES_WITH_INVEST_FLOW_REL, rule=_power_coupled
        )

    def _objective_expression(self):
        r"""
        Objective expression for storages with no investment.
        Note: This adds nothing as variable costs are already
        added in the Block :class:`SimpleFlowBlock`.
        """
        if not hasattr(self, "STORAGES"):
            return 0

        return 0


class GenericInvestmentStorageBlockPit(ScalarBlock):

    CONSTRAINT_GROUP = True

    def __init__(self, *args,**kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """ """
        m = self.parent_block()
        if group is None:
            return None

        # ########################## SETS #####################################

        self.INVESTSTORAGES = Set(initialize=[n for n in group])

        self.CONVEX_INVESTSTORAGES = Set(
            initialize=[n for n in group if n.investment.nonconvex is False]
        )

        self.NON_CONVEX_INVESTSTORAGES = Set(
            initialize=[n for n in group if n.investment.nonconvex is True]
        )

        self.INVESTSTORAGES_BALANCED = Set(
            initialize=[n for n in group if n.balanced is True]
        )

        self.INVESTSTORAGES_NO_INIT_CONTENT = Set(
            initialize=[n for n in group if n.initial_storage_level is None]
        )

        self.INVESTSTORAGES_INIT_CONTENT = Set(
            initialize=[
                n for n in group if n.initial_storage_level is not None
            ]
        )

        self.INVEST_REL_CAP_IN = Set(
            initialize=[
                n
                for n in group
                if n.invest_relation_input_capacity is not None
            ]
        )

        self.INVEST_REL_CAP_OUT = Set(
            initialize=[
                n
                for n in group
                if n.invest_relation_output_capacity is not None
            ]
        )

        self.INVEST_REL_IN_OUT = Set(
            initialize=[
                n for n in group if n.invest_relation_input_output is not None
            ]
        )

        # The storage content is a non-negative variable, therefore it makes no
        # sense to create an additional constraint if the lower bound is zero
        # for all time steps.
        self.MIN_INVESTSTORAGES = Set(
            initialize=[
                n
                for n in group
                if sum([n.min_storage_level[t] for t in m.TIMESTEPS]) > 0
            ]
        )

        # ######################### Variables  ################################
        self.storage_content = Var(
            self.INVESTSTORAGES, m.TIMESTEPS, within=NonNegativeReals
        )

        self.storage_height = Var(self.INVESTSTORAGES, within=NonNegativeReals)

        self.storage_height_reciprocal = Var(self.INVESTSTORAGES, within=NonNegativeReals)

        def _storage_investvar_bound_rule(block, n):
            """
            Rule definition to bound the invested storage capacity `invest`.
            """
            if n in self.CONVEX_INVESTSTORAGES:
                return n.investment.minimum, n.investment.maximum
            elif n in self.NON_CONVEX_INVESTSTORAGES:
                return 0, n.investment.maximum

        self.invest = Var(
            self.INVESTSTORAGES,
            within=NonNegativeReals,
            bounds=_storage_investvar_bound_rule,
        )

        self.init_content = Var(self.INVESTSTORAGES, within=NonNegativeReals)

        # create status variable for a non-convex investment storage
        self.invest_status = Var(self.NON_CONVEX_INVESTSTORAGES, within=Binary)

        # ######################### CONSTRAINTS ###############################
        i = {n: [i for i in n.inputs][0] for n in group}
        o = {n: [o for o in n.outputs][0] for n in group}

        reduced_timesteps = [x for x in m.TIMESTEPS if x > 0]

        def _inv_storage_init_content_max_rule(block, n):
            """Constraint for a variable initial storage capacity."""
            return (
                block.init_content[n]
                <= n.investment.existing + block.invest[n]
            )

        self.init_content_limit = Constraint(
            self.INVESTSTORAGES_NO_INIT_CONTENT,
            rule=_inv_storage_init_content_max_rule,
        )

        def _inv_storage_init_content_fix_rule(block, n):
            """Constraint for a fixed initial storage capacity."""
            return block.init_content[n] == n.initial_storage_level * (
                n.investment.existing + block.invest[n]
            )

        self.init_content_fix = Constraint(
            self.INVESTSTORAGES_INIT_CONTENT,
            rule=_inv_storage_init_content_fix_rule,
        )

        def _storage_balance_first_rule(block, n):
            """
            Rule definition for the storage balance of every storage n for the
            first time step.
            """
            expr = 0
            expr += block.storage_content[n, 0]
            expr += (
                -block.init_content[n]
                * (1 - n.loss_rate[0] * block.storage_height_reciprocal[n]) * m.timeincrement[0]
            )
            expr += (
                n.fixed_losses_relative[0]
                * block.storage_height_reciprocal[n]
                * (n.investment.existing + self.invest[n])
                * m.timeincrement[0]
            )
            expr += n.fixed_losses_absolute[0] * block.storage_height_reciprocal[n] * m.timeincrement[0]
            expr += (
                -m.flow[i[n], n, 0] * n.inflow_conversion_factor[0]
            ) * m.timeincrement[0]
            expr += (
                m.flow[n, o[n], 0] / n.outflow_conversion_factor[0]
            ) * m.timeincrement[0]
            return expr == 0

        self.balance_first = Constraint(
            self.INVESTSTORAGES, rule=_storage_balance_first_rule
        )

        def _storage_balance_rule(block, n, t):
            """
            Rule definition for the storage balance of every storage n for the
            every time step but the first.
            """
            expr = 0
            expr += block.storage_content[n, t]
            expr += (
                -block.storage_content[n, t - 1]
                * (1 - n.loss_rate[t] * block.storage_height_reciprocal[n]) * m.timeincrement[t]
            )
            expr += (
                n.fixed_losses_relative[t]
                * block.storage_height_reciprocal[n]
                * (n.investment.existing + self.invest[n])
                * m.timeincrement[t]
            )
            expr += n.fixed_losses_absolute[t] * block.storage_height_reciprocal[n] * m.timeincrement[t]
            expr += (
                -m.flow[i[n], n, t] * n.inflow_conversion_factor[t]
            ) * m.timeincrement[t]
            expr += (
                m.flow[n, o[n], t] / n.outflow_conversion_factor[t]
            ) * m.timeincrement[t]
            return expr == 0

        self.balance = Constraint(
            self.INVESTSTORAGES, reduced_timesteps, rule=_storage_balance_rule
        )

        def _balanced_storage_rule(block, n):
            return (
                block.storage_content[n, m.TIMESTEPS.at(-1)]
                == block.init_content[n]
            )

        self.balanced_cstr = Constraint(
            self.INVESTSTORAGES_BALANCED, rule=_balanced_storage_rule
        )

        def _storage_height_capacity_connection_rule(block, n):
            if n.investment.maximum * 3600 / ((n.temp_h-n.temp_c) * n.rho * n.c) <= 30000000:  # Litres
                return (
                    block.storage_height[n] == (0.000127839/(n.temp_h-n.temp_c)) *
                    (n.investment.existing + self.invest[n]) + 4.247933
                )
            else:
                return (
                    block.storage_height[n] == (0.0000357789/(n.temp_h - n.temp_c)) *
                    (n.investment.existing + self.invest[n]) + 8.1381764
                )

        self.height_cstr = Constraint(
            self.INVESTSTORAGES, rule=_storage_height_capacity_connection_rule
        )

        def _storage_height_reciprocal_rule(block, n):
            return block.storage_height_reciprocal[n] * block.storage_height[n] == 1

        self.storage_height_constraints = Constraint(
            self.INVESTSTORAGES, rule=_storage_height_reciprocal_rule
        )

        def _power_coupled(block, n):
            """
            Rule definition for constraint to connect the input power
            and output power
            """
            expr = (
                m.InvestmentFlowBlock.invest[n, o[n]]
                + m.flows[n, o[n]].investment.existing
            ) * n.invest_relation_input_output == (
                m.InvestmentFlowBlock.invest[i[n], n]
                + m.flows[i[n], n].investment.existing
            )
            return expr

        self.power_coupled = Constraint(
            self.INVEST_REL_IN_OUT, rule=_power_coupled
        )

        def _storage_capacity_inflow_invest_rule(block, n):
            """
            Rule definition of constraint connecting the inflow
            `InvestmentFlowBlock.invest of storage with invested capacity
            `invest` by nominal_storage_capacity__inflow_ratio
            """
            expr = (
                m.InvestmentFlowBlock.invest[i[n], n]
                + m.flows[i[n], n].investment.existing
            ) == (
                n.investment.existing + self.invest[n]
            ) * n.invest_relation_input_capacity
            return expr

        self.storage_capacity_inflow = Constraint(
            self.INVEST_REL_CAP_IN, rule=_storage_capacity_inflow_invest_rule
        )

        def _storage_capacity_outflow_invest_rule(block, n):
            """
            Rule definition of constraint connecting outflow
            `InvestmentFlowBlock.invest` of storage and invested capacity
            `invest` by nominal_storage_capacity__outflow_ratio
            """
            expr = (
                m.InvestmentFlowBlock.invest[n, o[n]]
                + m.flows[n, o[n]].investment.existing
            ) == (
                n.investment.existing + self.invest[n]
            ) * n.invest_relation_output_capacity
            return expr

        self.storage_capacity_outflow = Constraint(
            self.INVEST_REL_CAP_OUT, rule=_storage_capacity_outflow_invest_rule
        )

        def _max_storage_content_invest_rule(block, n, t):
            """
            Rule definition for upper bound constraint for the
            storage content.
            """
            expr = (
                self.storage_content[n, t]
                <= (n.investment.existing + self.invest[n])
                * n.max_storage_level[t]
            )
            return expr

        self.max_storage_content = Constraint(
            self.INVESTSTORAGES,
            m.TIMESTEPS,
            rule=_max_storage_content_invest_rule,
        )

        def _min_storage_content_invest_rule(block, n, t):
            """
            Rule definition of lower bound constraint for the
            storage content.
            """
            expr = (
                self.storage_content[n, t]
                >= (n.investment.existing + self.invest[n])
                * n.min_storage_level[t]
            )
            return expr

        # Set the lower bound of the storage content if the attribute exists
        self.min_storage_content = Constraint(
            self.MIN_INVESTSTORAGES,
            m.TIMESTEPS,
            rule=_min_storage_content_invest_rule,
        )

        def maximum_invest_limit(block, n):
            """
            Constraint for the maximal investment in non convex investment
            storage.
            """
            return (
                n.investment.maximum * self.invest_status[n] - self.invest[n]
            ) >= 0

        self.limit_max = Constraint(
            self.NON_CONVEX_INVESTSTORAGES, rule=maximum_invest_limit
        )

        def smallest_invest(block, n):
            """
            Constraint for the minimal investment in non convex investment
            storage if the invest is greater than 0. So the invest variable
            can be either 0 or greater than the minimum.
            """
            return (
                self.invest[n] - (n.investment.minimum * self.invest_status[n])
                >= 0
            )

        self.limit_min = Constraint(
            self.NON_CONVEX_INVESTSTORAGES, rule=smallest_invest
        )

    def _objective_expression(self):
        """Objective expression with fixed and investement costs."""
        if not hasattr(self, "INVESTSTORAGES"):
            return 0

        investment_costs = 0

        for n in self.CONVEX_INVESTSTORAGES:
            investment_costs += self.invest[n] * n.investment.ep_costs
        for n in self.NON_CONVEX_INVESTSTORAGES:
            investment_costs += (
                self.invest[n] * n.investment.ep_costs
                + self.invest_status[n] * n.investment.offset
            )
        self.investment_costs = Expression(expr=investment_costs)

        return investment_costs



def labelDictGenerator(numBuildings, labels, optimType, mergedLinks):
    base = {"electricityLink":"elLink", "shLink":"shLink", "dhwLink":"dhwLink", "naturalGasResource":"natGas", "naturalGasBus":"natGas", "qSource":"qSource", "heatSourceBus":"hSB", "gridBus":"grid", "pv":"pv", "electricityResource":"grid", "gridElectricity":"grid", "GasBoiler":"gasBoiler",
    "CHP":"CHP", "electricityBus":"prodEl", "electricityProdBus":"localEl", "producedElectricity":"prodEl", "electricitySource":"localEl", "electricalStorage":"Bat", "excesselectricityBus":"exEl",
    "excessshDemandBus":"exSh", "electricityInBus":"usedEl", "HP":"HP", "GWHP":"GWHP", "GWHP35":"GWHP35", "GWHP60":"GWHP60", "solarCollector":"solar", "solarConnectBus":"solar","heat_solarCollector":"solar", "excess_solarheat":"exSolar",
    "shSource":"prodSH","shSourceBus":"prodSH", "spaceHeatingBus":"shBus", "spaceHeating":"shBus", "shStorage":"shStor", "shDemandBus":"shBus", "dhwStorageBus":"dhwStor", "dhwStorage":"dhwStor", "domesticHotWaterBus":"dhwBus",
    "domesticHotWater":"dhwBus", "dhwDemandBus":"dhwBus", "electricityDemand":"Q_el", "emobilityDemand":"Q_mob", "spaceHeatingDemand":"Q_sh", "domesticHotWaterDemand":"Q_dhw", "excessshSourceBus":"exSh",
    "ElectricRod":"ElectricRod"}
    if not mergedLinks and optimType == 'group':
        base['electricityInBus'] = "usedEl"
        base['spaceHeatingBus'] = "usedSH"
        base['domesticHotWaterBus'] = "prodDHW"
        if labels != 'default':
            base["electricityInBus"] = labels["usedEl"]
            base["spaceHeatingBus"] = labels["usedSH"]
            base["spaceHeating"] = labels["shBus"]
            base["shDemandBus"] = labels["shBus"]
            base["domesticHotWaterBus"] = labels["prodDHW"]
            base['domesticHotWater'] = labels["dhwBus"]
            base['dhwDemandBus'] = labels["dhwBus"]
    if labels != 'default':
        if "elBus" in labels and ((mergedLinks and optimType=='group') or optimType=='indiv'): base["electricityLink"] = labels["elBus"]
        if "elBus" in labels and not mergedLinks and optimType=='group': base["electricityLink"] = labels["elBus"] + " Link"
        if "shBus" in labels and ((mergedLinks and optimType=='group') or optimType=='indiv'): base["shLink"] = labels["shBus"]
        if "shBus" in labels and not mergedLinks and optimType=='group': base["shLink"] = labels["shBus"] + " Link"
        if "dhwBus" in labels and ((mergedLinks and optimType=='group') or optimType=='indiv'): base["dhwLink"] = labels["dhwBus"]
        if "dhwBus" in labels and not mergedLinks and optimType=='group': base["dhwLink"] = labels["dhwBus"] + " Link"
        if "naturalGas" in labels:
            base["naturalGasResource"] = labels["naturalGas"]
            base["naturalGasBus"] = labels["naturalGas"]
        if "qSource" in labels:
            base["qSource"] = labels["qSource"]
            base["heatSourceBus"] = labels["hSB"]
        if "grid" in labels:
            base["gridBus"] = labels["grid"]
            base["electricityResource"] = labels["grid"]
            base["gridElectricity"] = labels["grid"]
        if "pv" in labels: base["pv"]=labels["pv"]
        if "gasBoiler" in labels: base["GasBoiler"]=labels["gasBoiler"]
        if "CHP" in labels:base["CHP"]=labels["CHP"]
        if "prodEl" in labels:
            base["electricityBus"]=labels["prodEl"]
            base["producedElectricity"] = labels["prodEl"]
        if "localEl" in labels:
            base["electricityProdBus"]=labels["localEl"]
            base["electricitySource"]=labels["localEl"]
        if "StorageEl" in labels:base["electricalStorage"]=labels["StorageEl"]
        if "excessEl" in labels:base["excesselectricityBus"]=labels["excessEl"]
        if "excessSh" in labels:
            base["excessshDemandBus"]=labels["excessSh"]
            base["excessshSourceBus"]=labels["excessSh"]
        if "elBus" in labels and (mergedLinks or optimType == 'indiv'):base["electricityInBus"]=labels["elBus"]
        if "HP" in labels:base["HP"]=labels["HP"]
        if "GWHP" in labels:
            base["GWHP"]=labels["GWHP"]
            base["GWHP35"]=labels["GWHP"]+'35'# for splitted GSHP
            base["GWHP60"] = labels["GWHP"]+'60'
        if "solarCollector" in labels:
            base["solarCollector"]=labels["solarCollector"]
            base["solarConnectBus"]=labels["solarCollector"]
            base["heat_solarCollector"]=labels["solarCollector"]
        if "excessSolarCollector" in labels: base["excess_solarheat"]=labels["excessSolarCollector"]
        if "prodSH" in labels:
            base["shSource"]=labels["prodSH"]
            base["shSourceBus"]=labels["prodSH"]
        if "shBus" in labels and (mergedLinks or optimType == 'indiv'):
            base["spaceHeatingBus"]=labels["shBus"]
            base["spaceHeating"]=labels["shBus"]
            base["shDemandBus"] = labels["shBus"]
        if "StorageSh" in labels:base["shStorage"]=labels["StorageSh"]
        if "StorageDhw" in labels:
            base["dhwStorageBus"]=labels["StorageDhw"]
            base["dhwStorage"]=labels["StorageDhw"]
        if "dhwBus" in labels and (mergedLinks or optimType == 'indiv'):
            base["domesticHotWaterBus"]=labels["dhwBus"]
            base["domesticHotWater"]=labels["dhwBus"]
            base["dhwDemandBus"]=labels["dhwBus"]
        if "DemandEl" in labels:base["electricityDemand"]=labels["DemandEl"]
        if "DemandMob" in labels:base["emobilityDemand"]=labels["DemandMob"]
        if "DemandSh" in labels:base["spaceHeatingDemand"]=labels["DemandSh"]
        if "DemandDhw" in labels:base["domesticHotWaterDemand"]=labels["DemandDhw"]
        if "ElectricRod" in labels:base["ElectricRod"]=labels["ElectricRod"]

    labelDict = {}

    for b in range(1,numBuildings+1):
        for key in base:
            if ("grid" in base[key] or "Grid" in base[key]) and mergedLinks:    # combine grid bus for merged links
                value = base[key]
                key = key + "__Building" + str(b)
            elif mergedLinks and (all(v not in key for v in ["electricityLink", "shLink", "dhwLink", "electricityInBus", "domesticHotWater", "spaceHeatingBus", "domesticHotWaterBus", "dhwDemandBus", "spaceHeating", "shDemandBus"])
            or any(v in key for v in ["spaceHeatingDemand", 'domesticHotWaterDemand'])):            # append suffix for all values except links
                value = base[key]+"_B"+str(b)
                key = key+"__Building"+str(b)
            elif ((not mergedLinks and optimType == "group") or optimType == "indiv") and all(v not in key for v in ["electricityLink", "shLink", "dhwLink"]):
                value = base[key] + "_B" + str(b)
                key = key + "__Building" + str(b)
            else:
                value = base[key]
                if mergedLinks and any(v in key for v in ["electricityInBus", "domesticHotWater", "spaceHeatingBus", "domesticHotWaterBus", "dhwDemandBus", "spaceHeating", "shDemandBus"]):            # append suffix for all values except links
                    key = key+"__Building"+str(b)
            labelDict[key] = value

    return labelDict

def positionDictGenerator(labels, optimType, mergedLinks):
    labelsList = ['natGas', 'grid', 'pv', 'CHP', 'gasBoiler', 'localEl', 'prodEl', 'elLink', 'shLink', 'dhwLink', 'Bat',
                  'usedEl', 'HP', 'GWHP', 'ElectricRod', 'solar', 'exSolar', 'prodSH', 'shStor', 'dhwStor', 'exEl',
                  'Q_el', 'Q_mob', 'Q_sh', 'Q_dhw', 'exSh', 'dhwBus', 'shBus', 'qSource', 'hSB']
    if not mergedLinks and optimType == 'group':
        labelsList.extend(['usedEl', 'usedSH', 'prodDHW'])
    if labels != 'default':
        if "elBus" in labels and ((mergedLinks and optimType=='group') or optimType=='indiv'): labelsList[7] = labels["elBus"]
        if "elBus" in labels and not mergedLinks and optimType == 'group': labelsList[7] = labels["elBus"] + " Link"
        if "shBus" in labels and ((mergedLinks and optimType=='group') or optimType=='indiv'): labelsList[8] = labels["shBus"]
        if "shBus" in labels and not mergedLinks and optimType == 'group': labelsList[8] = labels["shBus"] + " Link"
        if "dhwBus" in labels and ((mergedLinks and optimType=='group') or optimType=='indiv'): labelsList[9] = labels["dhwBus"]
        if "dhwBus" in labels and not mergedLinks and optimType == 'group': labelsList[9] = labels["dhwBus"] + " Link"
        if "naturalGas" in labels: labelsList[0] = labels["naturalGas"]
        if "grid" in labels: labelsList[1] = labels["grid"]
        if "pv" in labels: labelsList[2]=labels["pv"]
        if "gasBoiler" in labels: labelsList[4]=labels["gasBoiler"]
        if "CHP" in labels:labelsList[3]=labels["CHP"]
        if "prodEl" in labels: labelsList[6] = labels["prodEl"]
        if "localEl" in labels: labelsList[5]=labels["localEl"]
        if "StorageEl" in labels:labelsList[10]=labels["StorageEl"]
        if "excessEl" in labels:labelsList[20]=labels["excessEl"]
        if "excessSh" in labels:labelsList[25]=labels["excessSh"]
        if "elBus" in labels:labelsList[11]=labels["elBus"]
        if "HP" in labels:labelsList[12]=labels["HP"]
        if "GWHP" in labels:labelsList[13]=labels["GWHP"]
        if "solarCollector" in labels:labelsList[15]=labels["solarCollector"]
        if "excessSolarCollector" in labels: labelsList[16]=labels["excessSolarCollector"]
        if "prodSH" in labels:labelsList[17]=labels["prodSH"]
        if "shBus" in labels:labelsList[27] = labels["shBus"]
        if "StorageSh" in labels:labelsList[18]=labels["StorageSh"]
        if "StorageDhw" in labels:labelsList[19]=labels["StorageDhw"]
        if "dhwBus" in labels:labelsList[26]=labels["dhwBus"]
        if "DemandEl" in labels:labelsList[21]=labels["DemandEl"]
        if "DemandMob" in labels:labelsList[22]=labels["DemandMob"]
        if "DemandSh" in labels:labelsList[23]=labels["DemandSh"]
        if "DemandDhw" in labels:labelsList[24]=labels["DemandDhw"]
        if "hSB" in labels: labelsList[29] = labels["hSB"]
        if "qSource" in labels: labelsList[28] = labels["qSource"]
        if "ElectricRod" in labels:labelsList[14]=labels["ElectricRod"]
    positionDict = {
        labelsList[0]: [0.001, 0.65],  # X and Y positions should never be set to 0 or 1
        labelsList[1]: [0.001, 0.15],
        labelsList[2]: [0.001, 0.3],
        labelsList[3]: [0.1, 0.7],
        labelsList[4]: [0.1, 0.7],
        labelsList[5]: [0.15, 0.3],
        labelsList[6]: [0.3, 0.3],
        labelsList[10]: [0.2, 0.25],
        labelsList[11]: [0.4, 0.2],
        labelsList[12]: [0.55, 0.5],
        labelsList[13]: [0.55, 0.2],
        labelsList[13]+'35': [0.55, 0.2],
        labelsList[13]+'60': [0.55, 0.5],
        labelsList[14]: [0.6, 0.4],
        labelsList[15]: [0.6, 0.85],
        labelsList[16]: [0.7, 0.95],
        labelsList[17]: [0.65, 0.58],
        # "usedSH":	[0.75, 0.58],
        labelsList[18]: [0.68, 0.37],
        labelsList[19]: [0.65, 0.85],
        # "usedDHW": [0.8, 0.85],
        # "prodDHW": [0.75, 0.85],
        labelsList[20]: [0.999, 0.25],
        labelsList[21]: [0.999, 0.15],
        labelsList[22]: [0.999, 0.4],
        labelsList[23]: [0.999, 0.6],
        labelsList[24]: [0.999, 0.85],
        labelsList[25]: [0.999, 0.7],
        labelsList[26]: [0.8, 0.85],
        labelsList[27]: [0.8, 0.6]
    }

    if mergedLinks and optimType=='group':
        positionDict[labelsList[7]]= [0.4, 0.35]
        positionDict[labelsList[8]]= [0.8, 0.35]
        positionDict[labelsList[9]]= [0.8, 0.5]
    elif optimType!='indiv':
        #links
        positionDict[labelsList[7]] = [0.45, 0.35]
        positionDict[labelsList[8]] = [0.77, 0.35]
        positionDict[labelsList[9]] = [0.85, 0.5]
        # prod DHW, used SH and used El
        if labels != 'default':
            labelsList[28] = labels["usedEl"]
            labelsList[29] = labels["usedSH"]
            labelsList[30] = labels["prodDHW"]
        positionDict[labelsList[28]] = [0.52, 0.35]
        positionDict[labelsList[29]] = [0.75, 0.35]
        positionDict[labelsList[30]] = [0.7, 0.5]

    return positionDict

labelDict = {
    "electricityLink": "elLink",
    "electricityLink1_2": "elLink",
    "electricityLink1_3": "elLink",
    "electricityLink1_4": "elLink",
    "electricityLink1_5": "elLink",
    "electricityLink1_6": "elLink",
    "electricityLink2_3": "elLink",
    "electricityLink2_4": "elLink",
    "electricityLink2_5": "elLink",
    "electricityLink2_6": "elLink",
    "electricityLink3_4": "elLink",
    "electricityLink3_5": "elLink",
    "electricityLink3_6": "elLink",
    "electricityLink4_5": "elLink",
    "electricityLink4_6": "elLink",
    "electricityLink5_6": "elLink",
    "electricityLink2_1": "elLink",
    "electricityLink3_2": "elLink",
    "electricityLink3_1": "elLink",
    "electricityLink4_3": "elLink",
    "electricityLink4_2": "elLink",
    "electricityLink4_1": "elLink",
    "electricityLink5_4": "elLink",
    "electricityLink5_3": "elLink",
    "electricityLink5_2": "elLink",
    "electricityLink5_1": "elLink",
    "electricityLink6_5": "elLink",
    "electricityLink6_4": "elLink",
    "electricityLink6_3": "elLink",
    "electricityLink6_2": "elLink",
    "electricityLink6_1": "elLink",

    "shLink1_2": "shLink",
    "shLink1_3": "shLink",
    "shLink1_4": "shLink",
    "shLink1_5": "shLink",
    "shLink1_6": "shLink",
    "shLink2_3": "shLink",
    "shLink2_4": "shLink",
    "shLink2_5": "shLink",
    "shLink2_6": "shLink",
    "shLink3_4": "shLink",
    "shLink3_5": "shLink",
    "shLink3_6": "shLink",
    "shLink4_5": "shLink",
    "shLink4_6": "shLink",
    "shLink5_6": "shLink",
    "shLink2_1": "shLink",
    "shLink3_2": "shLink",
    "shLink3_1": "shLink",
    "shLink4_3": "shLink",
    "shLink4_2": "shLink",
    "shLink4_1": "shLink",
    "shLink5_1": "shLink",
    "shLink5_2": "shLink",
    "shLink5_3": "shLink",
    "shLink5_4": "shLink",
    "shLink6_5": "shLink",
    "shLink6_4": "shLink",
    "shLink6_3": "shLink",
    "shLink6_2": "shLink",
    "shLink6_1": "shLink",

    "dhwLink1_2": "dhwLink",
    "dhwLink1_3": "dhwLink",
    "dhwLink1_4": "dhwLink",
    "dhwLink1_5": "dhwLink",
    "dhwLink1_6": "dhwLink",
    "dhwLink2_3": "dhwLink",
    "dhwLink2_4": "dhwLink",
    "dhwLink2_5": "dhwLink",
    "dhwLink2_6": "dhwLink",
    "dhwLink3_4": "dhwLink",
    "dhwLink3_5": "dhwLink",
    "dhwLink3_6": "dhwLink",
    "dhwLink4_5": "dhwLink",
    "dhwLink4_6": "dhwLink",
    "dhwLink5_6": "dhwLink",
    "dhwLink2_1": "dhwLink",
    "dhwLink3_2": "dhwLink",
    "dhwLink3_1": "dhwLink",
    "dhwLink4_3": "dhwLink",
    "dhwLink4_2": "dhwLink",
    "dhwLink4_1": "dhwLink",
    "dhwLink5_1": "dhwLink",
    "dhwLink5_2": "dhwLink",
    "dhwLink5_3": "dhwLink",
    "dhwLink5_4": "dhwLink",
    "dhwLink6_5": "dhwLink",
    "dhwLink6_4": "dhwLink",
    "dhwLink6_3": "dhwLink",
    "dhwLink6_2": "dhwLink",
    "dhwLink6_1": "dhwLink",
    "naturalGasResource__Building1": "natGas_B1",
    "naturalGasBus__Building1": "natGas_B1",
    "qSource__Building1": "qSource_B1",
    "heatSourceBus__Building1": "hSB_B1",
    "gridBus__Building1": "grid_B1",
    "pv__Building1": "pv_B1",
    "electricityResource__Building1":"grid_B1",
    "gridElectricity__Building1": "grid_B1",
    "GasBoiler__Building1":"gasBoiler_B1",
    "CHP_SH__Building1": "CHP_B1",
    "CHP_DHW__Building1": "CHP_B1",
    "CHP__Building1": "CHP_B1",
    "electricityBus__Building1": "prodEl_B1",
    "electricityProdBus__Building1":"localEl_B1", #localEl is before the battery, prodEl after the battery but before the elLink
    "producedElectricity__Building1": "prodEl_B1",
    "electricitySource__Building1": "localEl_B1",
    "electricalStorage__Building1": "Bat_B1",
    "excesselectricityBus__Building1": "exEl_B1",
    "excessshDemandBus__Building1": "exSh_B1",
    "electricityInBus__Building1": "usedEl_B1",
    "HP_SH__Building1": "HP_B1",
    "HP_DHW__Building1": "HP_B1",
    "HP__Building1": "HP_B1",
    "GWHP_SH__Building1": "GWHP_B1",
    "GWHP_DHW__Building1": "GWHP_B1",
    "GWHP__Building1": "GWHP_B1",
    "solarCollector__Building1": "solar_B1",
    "solarConnectBus__Building1": "solar_B1",
    'heat_solarCollector__Building1': "solar_B1",
    'excess_solarheat__Building1':"exSolar_B1",
    'shSource__Building1'	: "prodSH_B1",
    'shSourceBus__Building1': "prodSH_B1",
    "spaceHeatingBus__Building1": "shBus_B1",
    "spaceHeating__Building1": "shBus_B1",
    "shStorage__Building1": "shStor_B1",
    "shDemandBus__Building1": "shBus_B1",
    "dhwStorageBus__Building1": "dhwStor_B1",
    "dhwStorage__Building1": "dhwStor_B1",
    "domesticHotWaterBus__Building1": "dhwBus_B1",
    'domesticHotWater__Building1': "dhwBus_B1",
    "dhwDemandBus__Building1": "dhwBus_B1",
    "electricityDemand__Building1": "Q_el_B1",
    "emobilityDemand__Building1": "Q_mob_B1",
    "spaceHeatingDemand__Building1": "Q_sh_B1",
    "domesticHotWaterDemand__Building1": "Q_dhw_B1",
    "excessshSourceBus__Building1": "exSh_B1",
    "ElectricRod__Building1": "ElectricRod_B1",

    "qSource__Building2": "qSource_B2",
    "heatSourceBus__Building2": "hSB_B2",
    "naturalGasResource__Building2": "natGas_B2",
    "naturalGasBus__Building2": "natGas_B2",
    "electricityResource__Building2": "grid_B2",
    "gridBus__Building2": "grid_B2",
    "pv__Building2": "pv_B2",
    "gridElectricity__Building2": "grid_B2",
    "GasBoiler__Building2":"gasBoiler_B2",
    "CHP_SH__Building2": "CHP_B2",
    "CHP_DHW__Building2": "CHP_B2",
    "CHP__Building2": "CHP_B2",
    "electricityBus__Building2": "prodEl_B2",
    "electricityProdBus__Building2": "localEl_B2",
    "producedElectricity__Building2": "prodEl_B2",
    "electricitySource__Building2": "localEl_B2",
    "electricalStorage__Building2": "Bat_B2",
    "excesselectricityBus__Building2": "exEl_B2",
    "excessshDemandBus__Building2": "exSh_B2",
    "electricityInBus__Building2": "usedEl_B2",
    "HP_SH__Building2": "HP_B2",
    "HP_DHW__Building2": "HP_B2",
    "HP__Building2": "HP_B2",
    "GWHP_SH__Building2": "GWHP_B2",
    "GWHP_DHW__Building2": "GWHP_B2",
    "GWHP__Building2": "GWHP_B2",
    "solarCollector__Building2": "solar_B2",
    "solarConnectBus__Building2": "solar_B2",
    'heat_solarCollector__Building2': "solar_B2",
    'excess_solarheat__Building2': "exSolar_B2",
    'shSource__Building2'	: "prodSH_B2",
    'shSourceBus__Building2': "prodSH_B2",
    "spaceHeatingBus__Building2": "shBus_B2",
    "spaceHeating__Building2": "shBus_B2",
    "shStorage__Building2": "shStor_B2",
    "shDemandBus__Building2": "shBus_B2",
    "dhwStorageBus__Building2": "dhwStor_B2",
    "dhwStorage__Building2": "dhwStor_B2",
    "domesticHotWaterBus__Building2": "dhwBus_B2",
    'domesticHotWater__Building2': "dhwBus_B2",
    "dhwDemandBus__Building2": "dhwBus_B2",
    "electricityDemand__Building2": "Q_el_B2",
    "emobilityDemand__Building2": "Q_mob_B2",
    "spaceHeatingDemand__Building2":        "Q_sh_B2",
    "domesticHotWaterDemand__Building2": "Q_dhw_B2",
    "excessshSourceBus__Building2": "exSh_B2",
    "ElectricRod__Building2": "ElectricRod_B2",
    "naturalGasResource__Building3": "natGas_B3",
    "naturalGasBus__Building3": "natGas_B3",
    "qSource__Building3": "qSource_B3",
    "heatSourceBus__Building3": "hSB_B3",
    "electricityResource__Building3": "grid_B3",
    "gridBus__Building3": "grid_B3",
    "GasBoiler__Building3": "gasBoiler_B3",
    "pv__Building3": "pv_B3",
    "gridElectricity__Building3": "grid_B3",
    "CHP_SH__Building3": "CHP_B3",
    "CHP_DHW__Building3": "CHP_B3",
    "CHP__Building3": "CHP_B3",
    "electricityBus__Building3": "prodEl_B3",
    "electricityProdBus__Building3": "localEl_B3",
    "producedElectricity__Building3": "prodEl_B3",
    "electricitySource__Building3": "localEl_B3",
    "electricalStorage__Building3": "Bat_B3",
    "excesselectricityBus__Building3": "exEl_B3",
    "excessshDemandBus__Building3": "exSh_B3",
    "electricityInBus__Building3": "usedEl_B3",
    "HP_SH__Building3": "HP_B3",
    "HP_DHW__Building3": "HP_B3",
    "HP__Building3": "HP_B3",
    "GWHP_SH__Building3": "GWHP_B3",
    "GWHP_DHW__Building3": "GWHP_B3",
    "GWHP__Building3": "GWHP_B3",
    "solarCollector__Building3": "solar_B3",
    "solarConnectBus__Building3": "solar_B3",
    'heat_solarCollector__Building3': "solar_B3",
    'excess_solarheat__Building3': "exSolar_B3",
    'shSource__Building3': "prodSH_B3",
    'shSourceBus__Building3': "prodSH_B3",
    "spaceHeatingBus__Building3": "shBus_B3",
    "spaceHeating__Building3": "shBus_B3",
    "shStorage__Building3": "shStor_B3",
    "shDemandBus__Building3": "shBus_B3",
    "dhwStorageBus__Building3": "dhwStor_B3",
    "dhwStorage__Building3": "dhwStor_B3",
    "domesticHotWaterBus__Building3": "dhwBus_B3",
    'domesticHotWater__Building3': "dhwBus_B3",
    "dhwDemandBus__Building3": "dhwBus_B3",
    "electricityDemand__Building3": "Q_el_B3",
    "emobilityDemand__Building3": "Q_mob_B3",
    "spaceHeatingDemand__Building3":        "Q_sh_B3",
    "domesticHotWaterDemand__Building3": "Q_dhw_B3",
    "excessshSourceBus__Building3": "exSh_B3",
    "ElectricRod__Building3": "ElectricRod_B3",
    "naturalGasResource__Building4": "natGas_B4",
    "naturalGasBus__Building4": "natGas_B4",
    "qSource__Building4": "qSource_B4",
    "heatSourceBus__Building4": "hSB_B4",
    "electricityResource__Building4": "grid_B4",
    "gridBus__Building4": "grid_B4",
    "pv__Building4": "pv_B4",
    "gridElectricity__Building4": "grid_B4",
    "GasBoiler__Building4":"gasBoiler_B4",
    "CHP_SH__Building4": "CHP_B4",
    "CHP_DHW__Building4": "CHP_B4",
    "CHP__Building4": "CHP_B4",
    "electricityBus__Building4": "prodEl_B4",
    "electricityProdBus__Building4": "localEl_B4",
    "producedElectricity__Building4": "prodEl_B4",
    "electricitySource__Building4": "localEl_B4",
    "electricalStorage__Building4": "Bat_B4",
    "excesselectricityBus__Building4": "exEl_B4",
    "excessshDemandBus__Building4": "exSh_B4",
    "electricityInBus__Building4": "usedEl_B4",
    "HP_SH__Building4": "HP_B4",
    "HP_DHW__Building4": "HP_B4",
    "HP__Building4": "HP_B4",
    "GWHP_SH__Building4": "GWHP_B4",
    "GWHP_DHW__Building4": "GWHP_B4",
    "GWHP__Building4": "GWHP_B4",
    "solarCollector__Building4": "solar_B4",
    "solarConnectBus__Building4": "solar_B4",
    'heat_solarCollector__Building4': "solar_B4",
    'excess_solarheat__Building4': "exSolar_B4",
    'shSource__Building4': "prodSH_B4",
    'shSourceBus__Building4': "prodSH_B4",
    "spaceHeatingBus__Building4": "shBus_B4",
    "spaceHeating__Building4": "shBus_B4",
    "shStorage__Building4": "shStor_B4",
    "shDemandBus__Building4": "shBus_B4",
    "dhwStorageBus__Building4": "dhwStor_B4",
    "dhwStorage__Building4": "dhwStor_B4",
    "domesticHotWaterBus__Building4": "dhwBus_B4",
    'domesticHotWater__Building4': "dhwBus_B4",
    "dhwDemandBus__Building4": "dhwBus_B4",
    "electricityDemand__Building4": "Q_el_B4",
    "emobilityDemand__Building4": "Q_mob_B4",
    "spaceHeatingDemand__Building4":        "Q_sh_B4",
    "domesticHotWaterDemand__Building4": "Q_dhw_B4",
    "excessshSourceBus__Building4": "exSh_B4",
    "ElectricRod__Building4": "ElectricRod_B4",
    "naturalGasResource__Building5": "natGas_B5",
    "naturalGasBus__Building5": "natGas_B5",
    "qSource__Building5": "qSource_B5",
    "heatSourceBus__Building5": "hSB_B5",
    "electricityResource__Building5": "grid_B5",
    "gridBus__Building5": "grid_B5",
    "pv__Building5": "pv_B5",
    "gridElectricity__Building5": "grid_B5",
    "GasBoiler__Building5":"gasBoiler_B5",
    "CHP_SH__Building5": "CHP_B5",
    "CHP_DHW__Building5": "CHP_B5",
    "CHP__Building5": "CHP_B5",
    "electricityBus__Building5": "prodEl_B5",
    "electricityProdBus__Building5": "localEl_B5",
    "producedElectricity__Building5": "prodEl_B5",
    "electricitySource__Building5": "prodEl_B5",
    "electricalStorage__Building5": "Bat_B5",
    "excesselectricityBus__Building5": "exEl_B5",
    "excessshDemandBus__Building5": "exSh_B5",
    "excessshSourceBus__Building5": "exSh_B5",
    "electricityInBus__Building5": "usedEl_B5",
    "HP_SH__Building5": "HP_B5",
    "HP_DHW__Building5": "HP_B5",
    "HP__Building5": "HP_B5",
    "GWHP_SH__Building5": "GWHP_B5",
    "GWHP_DHW__Building5": "GWHP_B5",
    "GWHP__Building5": "GWHP_B5",
    "solarCollector__Building5": "solar_B5",
    "solarConnectBus__Building5": "solar_B5",
    'heat_solarCollector__Building5': "solar_B5",
    'excess_solarheat__Building5': "exSolar_B5",
    'shSource__Building5'	: "prodSH_B5",
    'shSourceBus__Building5': "prodSH_B5",
    "spaceHeatingBus__Building5": "usedSH_B5",
    "spaceHeating__Building5": "Q_sh_B5",
    "shStorage__Building5": "shStor_B5",
    "shDemandBus__Building5": "Q_sh_B5",
    "dhwStorageBus__Building5": "dhwStor_B5",
    "dhwStorage__Building5": "dhwStor_B5",
    "domesticHotWaterBus__Building5": "prodDHW_B5",
    'domesticHotWater__Building5': "usedDHW_B5",
    "dhwDemandBus__Building5": "Q_dhw_B5",
    "electricityDemand__Building5": "Q_el_B5",
    "spaceHeatingDemand__Building5":        "Q_sh_B5",
    "domesticHotWaterDemand__Building5": "Q_dhw_B5",
    "ElectricRod__Building5": "ElectricRod_B5",
    "naturalGasResource__Building6": "natGas_B6",
    "naturalGasBus__Building6": "natGas_B6",
    "qSource__Building6": "qSource_B6",
    "heatSourceBus__Building6": "hSB_B6",
    "electricityResource__Building6": "grid_B6",
    "gridBus__Building6": "grid_B6",
    "pv__Building6": "pv_B6",
    "gridElectricity__Building6": "grid_B6",
    "GasBoiler__Building6":"gasBoiler_B6",
    "CHP_SH__Building6": "CHP_B6",
    "CHP_DHW__Building6": "CHP_B6",
    "CHP__Building6": "CHP_B6",
    "electricityBus__Building6": "prodEl_B6",
    "electricityProdBus__Building6": "localEl_B6",
    "producedElectricity__Building6": "prodEl_B6",
    "electricitySource__Building6": "prodEl_B6",
    "electricalStorage__Building6": "Bat_B6",
    "excesselectricityBus__Building6": "exEl_B6",
    "excessshDemandBus__Building6": "exSh_B6",
    "excessshSourceBus__Building6": "exSh_B6",
    "electricityInBus__Building6": "usedEl_B6",
    "HP_SH__Building6": "HP_B6",
    "HP_DHW__Building6": "HP_B6",
    "HP__Building6": "HP_B6",
    "GWHP_SH__Building6": "GWHP_B6",
    "GWHP_DHW__Building6": "GWHP_B6",
    "GWHP__Building6": "GWHP_B6",
    "solarCollector__Building6": "solar_B6",
    "solarConnectBus__Building6": "solar_B6",
    'heat_solarCollector__Building6': "solar_B6",
    'excess_solarheat__Building6': "exSolar_B6",
    'shSource__Building6'	: "prodSH_B6",
    'shSourceBus__Building6': "prodSH_B6",
    "spaceHeatingBus__Building6": "usedSH_B6",
    "spaceHeating__Building6": "Q_sh_B6",
    "shStorage__Building6": "shStor_B6",
    "shDemandBus__Building6": "Q_sh_B6",
    "dhwStorageBus__Building6": "dhwStor_B6",
    "dhwStorage__Building6": "dhwStor_B6",
    "domesticHotWaterBus__Building6": "prodDHW_B6",
    'domesticHotWater__Building6': "usedDHW_B6",
    "dhwDemandBus__Building6": "Q_dhw_B6",
    "electricityDemand__Building6": "Q_el_B6",
    "spaceHeatingDemand__Building6":        "Q_sh_B6",
    "domesticHotWaterDemand__Building6": "Q_dhw_B6",
    "ElectricRod__Building6": "ElectricRod_B6",
}

labelPositionDict={
    "natGas": [0.001, 0.65],
    "qSource":	[0.001, 0.65], #X and Y positions should never be set to 0 or 1
    "grid":	[0.001, 0.15],
    "pv":   [0.001, 0.3],
    "CHP":	[0.1, 0.7],
    "gasBoiler": [0.1, 0.7],
    "localEl":	[0.15, 0.3],
    "prodEl":[0.3,0.3],
    "elLink":	[0.4, 0.35],
    "shLink": [0.8, 0.35],
    "dhwLink": [0.8, 0.5],
	"Bat":	[0.2, 0.25],
    "usedEl":	[0.4, 0.2],
    "HP":	[0.55, 0.5],
    "GWHP":	[0.55, 0.2],
    "ElectricRod": [0.6, 0.4],
    "solar":	[0.6, 0.85],
    "exSolar": [0.7, 0.95],
    "prodSH": [0.65, 0.58],
    #"usedSH":	[0.75, 0.58],
    "shStor":	[0.68, 0.37],
    "dhwStor":	[0.65, 0.85],
    #"usedDHW": [0.8, 0.85],
    #"prodDHW": [0.75, 0.85],
    "exEl": [0.999, 0.25],
    "Q_el":	[0.999, 0.15],
    "Q_mob":[0.999, 0.4],
    "Q_sh":	[0.999, 0.6],
    "Q_dhw":[0.999, 0.85],
    "exSh": [0.999, 0.7],
    "dhwBus":[0.8, 0.85],
    "shBus": [0.8, 0.6]
	}

fullPositionDict={
	"naturalGas": [0.001, 0.75],
        "qSource":	[0.001, 0.75],
        "gridBus":	[0.001, 0.2],
        "pv":   [0.001, 0.3],
        "gridElect":	[0.1, 0.05],
        "electricityProdBus": [0.1, 0.2],
        "CHP":	[0.1, 0.6],
        "GasBoiler": [0.1, 0.7],
        "electricitySource": [0.15, 0.2],
        "electricityBus":	[0.2, 0.2],
        "producedElectricity":	[0.3, 0.25],
        "electricityLink":	[0.3, 0.35],
        "electricalStorage":	[0.3, 0.25],
        "excesselect":	[0.5, 0.05],
        "electricityInBus":	[0.5, 0.2],
        "HP":	[0.6, 0.5],
        "GWHP":	[0.6, 0.2],
        "ElectricRod": [0.6, 0.4],
        "heat_solarCollector": [0.4, 0.85],
        "solarConnect": [0.5, 0.85],
        "solarCollector_":	[0.6, 0.85],
        "excess_solarheat":[0.6,0.85],
        'shSource_': [0.63, 0.58],
        'shLink': [0.64, 0.58],
        'shSourceBus': [0.66, 0.58],
        "spaceHeatingBus":	[0.7, 0.58],
        "spaceHeating_":	[0.8, 0.6],
        "shStorage":	[0.8, 0.37],
        "shDemandBus":	[0.9, 0.6],
        "dhwStorageBus":	[0.7, 0.9],
        "dhwStorage_":	[0.8, 0.9],
        "domesticHotWater_":	[0.92, 0.9],
        "domesticHotWaterBus":	[0.9, 0.9],
        "dhwDemandBus":	[0.95, 0.9],
        "electricityDemand":	[0.999, 0.1],
        "spaceHeatingDemand_":	[0.999, 0.6],
        "domesticHotWaterDemand":	[0.999, 0.9]
        }
